import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
from minimod_opt.utils.exceptions import MissingOptimizationMethod
import functools
import geopandas as gpd
from matplotlib.patches import Ellipse

def plot_context(func):
    
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):   
        with mpl.rc_context(self.mpl_theme_dict):
            plot = func(self, *args, **kwargs)  
            return plot
    return wrapper

class Plotter:
    """This class is in charge of plotting the results of the `minimod`. 
    """     
               
    
    def __init__(self, model, dpi = 600):
        
        self.model = model
        self.dpi = dpi
        
        self.mpl_theme_dict = {
            'font.family' : 'serif',
            # 'yaxis.formatter' : mpl.ticker.FuncFormatter(lambda x, pos: "{:,}".format(x/1000) + 'K'),
            'savefig.dpi' : self.dpi
        }

        
    def _check_if_optimization(self):
        
        try:
            self.model.opt_df
        except NameError:
            raise MissingOptimizationMethod("Optimization Results have not been created. Please run the `fit` method.")
    
    def _define_axis_object(self, fig = None, ax = None):
        
        if ax is None:
            figure, axis = plt.subplots()
        else:
            figure, axis = fig, ax
        
        return figure, axis
    
    @plot_context         
    def _plot_process(self, figure = None, axis = None):
        
        self._check_if_optimization()
        
        fig, ax = self._define_axis_object(fig = figure, ax=axis)
        
        return fig, ax
    
    @plot_context       
    def _plot_lines(self, 
                    to_plot = None,
                    title = None,
                    xlabel = None,
                    ylabel = None,
                    figure = None, 
                    axis = None,
                    twin = False,
                    twin_ylabel = None,
                    save = None,
                    legend = None):
        
        fig, ax = self._plot_process(figure= figure, axis = axis)
        
        if not twin:
            
            (self.model.opt_df[to_plot]
            .groupby([self.model.time_col])
            .sum()
            .plot(title = title, 
                fig = fig, 
                ax=ax
                )
            )
            
        if twin:
            ax.plot(self.model.opt_df[to_plot[0]].groupby([self.model.time_col])
                    .sum(), 
                    color='red')
            
            ax2 = ax.twinx()
            ax2.plot(self.model.opt_df[to_plot[1]].groupby([self.model.time_col])
                     .sum(), 
                     color = 'blue')
            
            ax2.set_ylabel(twin_ylabel)
        
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        if twin:
            ax.legend(['Optimal Benefits'], loc= 'upper left')
            ax2.legend(['Optimal Costs'], loc = 'lower left')
        plt.tight_layout()
        
        if save is not None:
            plt.savefig(save)
        
        if twin:
            return fig, ax, ax2
        if not twin:
            return fig, ax
    
    @plot_context    
    def _plot_hist(self,
                   to_plot = None,
                   title = None,
                   xlabel = None,
                   ylabel = None,
                   figure = None,
                   axis = None, 
                   save = None):
        
        fig, ax = self._plot_process(figure =figure, axis = axis)
        
        (
            self.model.opt_df[to_plot]
            .groupby([self.model.time_col])
            .sum()
            .plot
            .bar(fig = fig, ax = ax, grid = True, width = 1)
         )
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        plt.tight_layout()
        
        if save is not None:
            plt.savefig(save)
        
        return fig, ax
    
    def _merge_shape_file(self, 
                          data = None,
                          map_df = None, 
                          merge_key = None):
        
        self._check_if_optimization()
        
        # Merge with `opt_df`
        
        df = (
            map_df
            .merge(
                data
                .reset_index(),
                left_on = [merge_key],
                right_on = [self.model.space_col]
            )
            # .set_index([self.model.intervention_col,
            #             self.model.space_col,
            #             self.model.time_col
            #             ])
        )
        
        return df
    
    def _shape_file_loc(self,
                        data = None,
                        intervention = None,
                        time = None):
        
        df = (data
              .loc[(intervention, slice(None), time ), :]
              )
        
        return df
    
    def _dissolve_interventions(self, 
                                data = None, 
                                aggfunc = None):
                
        return data.dissolve(by = [self.model.space_col, self.model.time_col], aggfunc = aggfunc)
        
    @plot_context
    def _plot_chloropleth(self, 
                          data = None,
                          intervention = None,
                          time = None,
                          optimum_interest = None,
                          map_df = None,
                          merge_key = None,
                          aggfunc = None,
                          title = None, 
                          ax = None,
                          save = None,
                          show_legend = True,
                          intervention_bubbles = False,
                          intervention_bubble_names = None,
                          **kwargs):
        
        if ax is None:
            fig, ax = plt.subplots()
                    
        df = (
            data
            .pipe(self._shape_file_loc, intervention = intervention, time = time)
            .groupby(self.model.space_col)
            .sum()
            .pipe(self._merge_shape_file, map_df = map_df, merge_key = merge_key)
        )
        
        
        df.plot(column = optimum_interest, 
                ax = ax, 
                legend = show_legend,
                **kwargs)
        
        if intervention_bubbles:
            
            if isinstance(intervention_bubbles, bool):
            
                df_bubble = (
                    df
                    .merge(self.model._intervention_list_space_time
                        .loc[(slice(None), time)], on = self.model.space_col)
                    .assign(centroid = lambda df: df['geometry'].centroid)

                )         
            
            elif intervention_bubbles == 'bau':
                df_bubble = (
                    df
                    .merge(self.model.bau_list
                        .loc[(slice(None), time)], on = self.model.space_col)
                    .assign(centroid = lambda df: df['geometry'].centroid)

                ) 
                
            df_bubble_final = (
                df_bubble
                .assign(**{k : df_bubble['int_appeared']
                        .str.extract(f'(?P<{k}>{k})') \
                            for k in intervention_bubble_names})
                .assign(bubble_name = lambda df: df[intervention_bubble_names]
                        .fillna('')
                        .apply(lambda row: '\n'.join(row), axis=1))
            )
            
            df_bubble_final.apply(lambda x: ax.annotate(text = x.bubble_name,
                                                        xy = x.centroid.coords[0],
                                                        size = 7,
                                                        color ='red'), 
                                  axis=1)
      
        ax.set_title(title)
        ax.set_xticklabels([])
        ax.set_xticks([])
        ax.set_yticklabels([])
        ax.set_yticks([])
        plt.tight_layout()
        
        if save is not None:
            plt.savefig(save)
            
        return ax        

    @plot_context        
    def _plot_multi_chloropleth(self, 
                                data = None,
                                time = None, 
                                intervention = None,
                                optimum_interest = None,
                                map_df = None,
                                merge_key = None,
                                aggfunc = None,
                                title = None,
                                axs = None,
                                save = None,
                                **kwargs):
        
        if time is None:
            time = (self.model.opt_df
                 .index
                 .get_level_values(level = self.model.time_col)
                 .unique()
                 .values
                 )
        
        mod = len(time) % 2
        
        if axs is None:
            fig, axs = plt.subplots(nrows = int(len(time)/2) + mod , 
                                    ncols=2,
                                    figsize = (10,20))
        
        
        for ax, plot in zip(np.array(axs).flatten(), time):
            
            ax = self._plot_chloropleth(data = data,
                                   intervention = intervention,
                                   time = plot,
                                   optimum_interest = optimum_interest,
                                   map_df = map_df, 
                                   merge_key = merge_key, 
                                   aggfunc = aggfunc,
                                   ax = ax,
                                   title = f"T = {plot}", 
                                   **kwargs)
            ax.set_xticklabels([])
            ax.set_xticks([])
            ax.set_yticklabels([])
            ax.set_yticks([])
            ax.set_title(title)
            
        plt.tight_layout()
        plt.savefig(save)
        
        return axs
    
    @plot_context
    def _plot_sim_hist(self, 
                       data,
                       benefit_col = None, 
                       cost_col = None,
                       benefit_title = None,
                       cost_title = None,
                       save = None):
        
        fig, ax = plt.subplots(1,2, figsize=(12,6))
        
        data[benefit_col].hist(ax = ax[0])
        data[cost_col].hist(ax = ax[1])
        
        ax[0].set_title(benefit_title)
        ax[1].set_title(cost_title)
        fig.suptitle("Simulation Distributions")
        
        if save is not None:
            plt.savefig(save, dpi=300)
        
        return fig, ax
    
    @plot_context
    def _plot_grouped_bar(self,
                          intervention_col = None,
                          space_col = None,
                          col_of_interest = None,
                          ylabel = None,
                          intervention_subset = None,
                          save = None,
                          ):
        
        plot = (
            self.model.opt_df
            [col_of_interest]
            .to_frame()
            .loc[(intervention_subset, slice(None), slice(None)), :]
            .groupby([intervention_col, space_col])
            .sum()
            .reset_index()
            .pivot(index = space_col, columns = intervention_col, values = col_of_interest)
            .plot
            .bar(figsize = (12,6))
        )
        
        plot.legend(bbox_to_anchor = (1,1))
        
        if save is not None:
            plt.savefig(save)
    
    @plot_context
    def _plot_chloropleth_getter(self, time):
                
        _plotter = {
            'single' : self._plot_chloropleth,
            'multi' : self._plot_multi_chloropleth
        }
        
        if isinstance(time, int):
            number = 'single'
        else:
            number = 'multi'

            
        return _plotter.get(number)


    


        
        
        
        
        
        
        
        
            
        
        
        
    



# %%
from typing import Callable
import pandas as pd
import math
import  numpy as np
from .data import CostDataProcessor
import scipy.optimize as opt
        
class PremixCostCalculator:
    
    def __init__(self, 
                 data: CostDataProcessor = None,
                 upcharge: int = None,
                 excipient_price: float= None):
        
        self.upcharge = upcharge if not None else 1
        self.excipient_price = excipient_price if not None else 1.5
        
        self.data = data
        
        self._amt_fort = data.amt_fort
        self.price = data.price
        
        self.nutrient_subtotal = self.amt_fort.sum()
        
    @property
    def amt_fort(self):
        return self._amt_fort
    
    @amt_fort.setter
    def amt_fort(self, value):
        if len(value) != self._amt_fort.shape[0]:
            raise ValueError("New value's length doesn't equal length of amt_fort")
            
        self._amt_fort.loc[:] = value
        
    def reset_amt_fort(self):
        
        self._amt_fort = self._d.amt_fort
        
    @property
    def excipient(self):
        return self.addition_rate - self.nutrient_subtotal
        
    @property
    def excipient_prop(self):
        return (self.addition_rate - self.nutrient_subtotal)/self.addition_rate
    
    @property
    def excipient_cost(self):
        return self.excipient_prop * self.excipient_price
    
    @property
    def addition_rate(self):
        return int(math.ceil((self.nutrient_subtotal*1.1)/50.0)) * 50
    
    @property 
    def nutrient_total(self): 
        return self.cost_premix.sum() + self.excipient_cost
    
    @property
    def total_cost(self):
        return self.nutrient_total + self.upcharge
    
    @property
    def prop_fort(self):
        return self.amt_fort/self.addition_rate
    
    @property
    def cost_premix(self):
        return self.prop_fort*self.price
    
    def total_vehicle(self, fmt = 'kg'):
        
        if fmt == 'kg':
            div = 1_000_000
        elif fmt == 'mt':
            div = 1_000
        
        return self.nutrient_total*(self.addition_rate/div)
    
    @property
    def total_cost_mt_vehicle(self):
        return (self.total_cost*self.addition_rate)/1_000
    
    def premix_cost_summary(self):
        cost_summary = {
            'Nutrients Total ($/kg)' : self.nutrient_total,
            'Upcharge ($/kg)' : self.upcharge,
            'Total Cost ($/kg vehicle)' : self.total_cost,
            'Total ($/kg vehicle)' : self.total_vehicle(fmt = 'kg'),
            'Total ($/MT vehicle)' : self.total_vehicle(fmt = 'mt'),
            'Total Cost per kg Premix' : self.total_cost,
            'Total Cost per MT of Vehicle' : self.total_cost_mt_vehicle
        }

        return pd.DataFrame(data = cost_summary.values(), 
                            index = cost_summary.keys()) 
        
    def recalculate_total_costs(self):
        
        self.amt_fort
        self.nutrient_subtotal
        self.addition_rate
        self.prop_fort
        self.price
        self.excipient_prop 
        self.excipient_price
        self.cost_premix.sum() 
        self.excipient_cost
        self.nutrient_total
        self.upcharge
        
        return self.total_cost
    
    def line_fit(self, nutrient : str, compound : str, func : Callable=None):
        """Fits benefits data to `func`     

        Args:
            func (Callable, optional): Function to fit data to. If None, line is chosen. Defaults to None.
        """        
        
        if func is None:
            func = lambda x, m, b: (m*x) + b
            
        y = []
        x = np.linspace(1, 20000, 10)
            
        for i in x:
            self.amt_fort.loc[(nutrient, compound)] = i
                        
            y.append(self.recalculate_total_costs())
                    
        popt, pcov = opt.curve_fit(func, x, y)
        
        def f(x):
            return func(x, *popt)
        
        return f, popt, pcov
            
if __name__ == '__main__':

    import matplotlib.pyplot as plt

    # Load data from github
    df = pd.read_csv("example/cost_data.csv")
# Check if we change Iron

    fortificants = ['Micronized ferric pyrophosphate',
    'Retinyl Palmitate- 250,000 IU/g (dry)',
    'Zinc Oxide',
    'Vit. B-12 0.1% WS',
    'Folic Acid'
    ]
    
    d = CostDataProcessor(
        data = df,
        vehicle = 'Bouillon',
        fortificant_col = 'compound',
        fortificant=fortificants,
        activity_col = 'fort_prop',
        overage_col = 'fort_over'
    )

    p = PremixCostCalculator(
        data = d
    )

    f, params, cov = p.line_fit()

    x = np.linspace(1, 20000, 10)
    plt.plot(x, f(x))


# %%



import pandas
import mip

class BAUConstraintCreator:
    
    def __init__(self):
        pass
    
    def bau_df(self, data:pandas.DataFrame, constraint:str, discounted_variable:str = None) -> pandas.DataFrame:
        """A dataframe of costs and benefits for the BAU (business as usual) intervention

        Args:
            data (pandas.DataFrame): input data
            constraint (str): name of dataframe BAU column
            discounted_variable (str, optional): outputs a series of the variable of interest. Defaults to None.

        Returns:
            pandas.DataFrame
        """
     
        if discounted_variable is None:
            discounted_variable = data.columns
        
        df = (data
         .loc[(constraint, 
               slice(None), 
               slice(None)),:][discounted_variable]
         )
                                
        return df
        
    
    def create_bau_constraint(self, data:pandas.DataFrame, constraint:str, discounted_variable:str, over:str = None)->pandas.DataFrame:
        """This function sums the values of each column in the given dataframe. 
            If the option `over' is provided, the function sums across groups as well

        Args:
            data (pandas.DataFrame): input data
            constraint (str): name of dataframe's column with information BAU
            discounted_variable (str): column of interest.
            over (str, optional): name of dataframe's column  with attribute used to group data by (e.g., time, region). Defaults to None.

        Returns:
            pandas.DataFrame: dataframe with the sum of values for each column-group
        """

      
        if over is None:
            minimum_constraint = (
                self.bau_df(data, constraint, discounted_variable)
                .sum()
            )
        else:
            minimum_constraint = (
                self.bau_df(data, constraint, discounted_variable)
                .groupby(over)
                .sum()
            )
        
        return minimum_constraint
    



from minimod_opt.base.basesolver import BaseSolver
from minimod_opt.utils.exceptions import NotPandasDataframe, MissingColumn
from minimod_opt.utils.summary import OptimizationSummary

import pandas as pd
import mip
import numpy as np

class BenefitSolver(BaseSolver):
    """This is the solver for benefit maximization. This fitter solves the problem of maximizing effective coverage subject to a cost constraint. This problem solved by this fitter is:
    
        $$
        \max \sum_{k} \sum_{t} Y_{k,t} \sum_{j} \frac{EfCvg_{k,j,t}}{(1+r)^t} + \sum_{k} \sum_{j} \sum_{t} X_{k,j,t} \frac{EfCvg_{k,j,t}}{(1+r)^t}
        s.t. 
        \sum_{t} Y_{k,t} \sum_{j} \frac{TC_{k,j,t}}{(1+r)^t} + \sum_{k} \sum_{j} \sum_{t} X_{k,j,t} \frac{TC_{k,j,t}}{(1+r)^t} \leq TF
        $$
        
    Arguments:
        data {pandas dataframe} -- A pandas dataframe with columns for each time period and a column for regions or whatever spatial dimension is used. The rows for each time period should give the benefits of each intervention at time t, and space j
        
        |k     | j   |t   | benefits/costs   |
        |------|-----|----|------------------|
        |maize |north|0   | 100              |
        |maize |south|0   | 50               |
        |maize |east |0   | 30               |
        |maize |west |0   | 20               |
        
                
    """

    def __init__(self, 
                  total_funds = 35821703,
                  **kwargs):
        
        super().__init__(sense = mip.MAXIMIZE, **kwargs)     

        if total_funds is not None:
            self.total_funds = total_funds
        else:
            raise Exception("No total funds constraint specified.")


    def _objective(self, model, x, **kwargs):
        
        benefit = self._df[self._benefit_col]     
        # Discounted costs
        
        model.objective = self._discounted_sum_all(benefit)
        
        
    def _constraint(self, model, x, **kwargs):
        
        cost = self._df[self._cost_col]
        
        ## Make benefits constraint be at least as large as the one from the minimum benefit intervention

        
        model += self._discounted_sum_all(cost) <= self.total_funds
        
        ## Also add constraint that only allows one intervention in a time period and region
        
    def fit(self, extra_const = None):
        return self._fit(extra_const = extra_const)
    
    def report(self):
        
        s = OptimizationSummary(self)
        
        super().report()
        
        total_costs = self.model.objective_bound
        total_benefits = self.model.objective_value

        results = [
            ('Total Funds', self.total_funds),
            ("Total Cost", total_costs ),
            ("Total Coverage", total_benefits)
        ]
        
        s.print_generic(results)
        
        s.print_ratio(name = "Coverage per Cost",
                      num = total_benefits,
                      denom = total_costs)
        
        s.print_grouper(name = "Total Cost and Coverage over Time",
                        data = self.opt_df,
                        style = 'markdown')


        
        
    
    
    
    



        


from tabulate import tabulate
import pandas as pd
import docx
import progressbar
import numpy as np

class PreOptimizationDataSummary:
    
    def __init__(self, 
                data, 
                benefit_col, 
                cost_col, 
                intervention_col, 
                space_col, 
                time_col, 
                benefit_title,
                cost_title = "Costs ($)",
                intervention_title = 'Intervention',
                intervention_subset = None,
                intervention_subset_titles = None,
                bau_intervention = None,
                bau_title = None,
                space_title = 'Regions',
                time_title = 'Time'):
        
        self.data = data
        self.benefit_col = benefit_col
        self.cost_col = cost_col
        self.intervention_col = intervention_col
        self.space_col = space_col
        self.time_col = time_col
        self.benefit_title = benefit_title
        self.cost_title = cost_title
        self.intervention_title = intervention_title
        self.space_title = space_title
        self.time_title = time_title
        self.intervention_subset = intervention_subset
        self.intervention_subset_titles = intervention_subset_titles
        self.bau_intervention = bau_intervention
        self.bau_title = bau_title
    
    def _rename_variables(self, data):
        
        df = (
            data
            .reset_index()
            .rename({
                self.benefit_col : self.benefit_title,
                self.cost_col : self.cost_title
            }, axis = 'columns')
        )
        
        if self.time_col in data.columns:
            df = df.rename({self.time_col : self.time_title}, axis = 'columns')
        if self.intervention_col in data.columns:
            df = df.rename({self.intervention_col : self.intervention_title}, axis = 'columns')
        if self.space_col in data.columns:
            df = df.rename({self.space_col : self.space_title}, axis = 'columns')
            
        return df
    
    def _rename_intervention_subset(self, data):
        
        df = (
            data
            .replace(to_replace = {self.intervention_col : self.intervention_subset_titles})
        )
        
        return df
    
    def _create_national_intervention(self, data):
        
        subset_data = data[[self.intervention_col, self.space_col, self.time_col, self.benefit_col, self.cost_col]]
        
        national_sums = subset_data.groupby([self.intervention_col, self.time_col]).sum()
        
        national_sums[self.space_col] = 'National'
        
        df = (
            subset_data
            .set_index([self.intervention_col, self.space_col, self.time_col])
            .append(
                national_sums
                .reset_index()
                .set_index([self.intervention_col, self.space_col, self.time_col])
            )
            )
        
        return df
    
    def _construct_variables(self, data, variables_of_interest):
        
        for var in variables_of_interest:
            
            if variables_of_interest[var] == 'cost_per_benefit':
                
                variables_of_interest[var] = lambda df: df[self.cost_col]/df[self.benefit_col]
            
            df = data.assign( var = variables_of_interest[var]).round({'var' : 2}).rename({'var' : var}, axis = 'columns')
            
        return df
    
    def _word_table(self, data, save_path):
        """Makes a word table out of a dataframe.
        Borrowed from stackoverflow:
        https://stackoverflow.com/questions/40596518/writing-a-python-pandas-dataframe-to-word-document
        """
        
        doc = docx.Document()
        
        table = doc.add_table(data.shape[0] + 1, data.shape[1])
        
        for j in range(data.shape[-1]):
            table.cell(0,j).text = data.columns[j]
        
        # add the rest of the data frame
        for i in progressbar.progressbar(range(data.shape[0])):
            for j in range(data.shape[-1]):
                table.cell(i+1,j).text = str(data.values[i,j])
                
        doc.save(save_path)
        print(f"Docx table saved in {save_path}")
                
    
    def _style(self, data, style = None, save_path = None):
        
        if style == 'html':
            return data.to_html()
        elif style == 'markdown':
            return data.to_markdown()
        elif style == 'latex':
            return data.to_latex()
        elif style == 'csv':
            data.reset_index().to_csv(save_path)
            print(f"CSV saved in {save_path}")
        elif style == 'word':
            self._word_table(data = data, save_path = save_path)
            print(f"Saved to {save_path}")
            
    def _num_formatter(self, data):
        
        data[self.benefit_col] = data[self.benefit_col].map("{:,.0f}".format)
        data[self.cost_col] = data[self.cost_col].map("{:,.0f}".format)
        
        return data
        
        
    def summary_table(self, 
                      variables_of_interest, 
                      grouping = None,
                      intervention_specific = True, 
                      style = None, 
                      save_path = 'table.docx'):
        
        grouper = []
        
        if intervention_specific:
            grouper.append(self.intervention_col)
        
        if grouping == 'over_space':
            grouper.append(self.space_col)
        elif grouping == 'over_time':
            grouper.append(self.time_col)
            
    
        # Get unique values of regions
        regions_list = (self.data
                        .reset_index(drop=True)
                        [self.space_col]
                        .unique()
                        .tolist()
                        )
        regions_list.append('National')
        
                    
        region_cat = pd.CategoricalDtype(categories = regions_list, ordered=True)
        
        intervention_list = self.intervention_subset.copy()
        intervention_list.append(self.bau_intervention)
        
        intervention_cat = pd.CategoricalDtype(list(self.intervention_subset_titles.values()), ordered=True)
        
        df = (
            self.data
            .pipe(self._create_national_intervention)
            .groupby(grouper).sum()
            .reset_index()
            .loc[lambda df: df[self.intervention_col].isin(intervention_list)]
            .pipe(self._construct_variables, variables_of_interest)
            .pipe(self._num_formatter)
            .pipe(self._rename_intervention_subset)
            .pipe(self._rename_variables)
            .drop(columns = 'index')
            )
        
        df[self.intervention_title] = df[self.intervention_title].astype(intervention_cat)
        if grouping == 'over_space':
            df[self.space_title] = df[self.space_title].astype(region_cat)
            df = df.sort_values([self.intervention_title, self.space_title])
            
        if style is not None:
            self._style(data = df, style=style, save_path=save_path)
        
        def color_cells(val, color):
            """
            Takes a scalar and returns a string with
            the css property `'color: red'` for negative
            strings, black otherwise.
            """
            return color
        
        return df.style.background_gradient()
        
class OptimizationSummary:

    def __init__(self, model, table_fmt = "psql", decimals = 3, **kwargs):

        
        self.model = model
        self.table_fmt = table_fmt
        self.decimals = decimals

    def _print_specific_style(self, style, data):
        if style == "html":
            return self.html_print(data)
        elif style == "markdown":
            return self.markdown_print(data)
        elif style == "latex":
            return self.latex_print(data)
        elif style == 'pandas':
            return self.pandas_show(data)
        else:
            raise ValueError("style not available.")

    def html_print(self, data):
        return print(data.to_html())

    def markdown_print(self, data):
        return print(data.to_markdown())

    def latex_print(self, data):
        return print(data.to_latex())

    def pandas_show(self, data):
        return print(data)

    def _group_summarizer(self,
                          data = None,
                          intervention_specific=None,
                          over_time=True,
                          across_space=False,
                          style=None):

        if style is None:
            print("[Note]: style not specified, printing as pandas dataframe")
            print_style = self.pandas_show
        if style is not None:
            print_style = self._print_specific_style

        grouper = []

        if intervention_specific is None:
            intervention_specific = slice(None)

        if across_space:
            grouper.append(self.model.space_col)

        if over_time:
            grouper.append(self.model.time_col)

        summary_data = (data
                        .loc[(intervention_specific, slice(None), slice(None)), :]
                        .groupby(grouper)
                        .sum())

        return print_style(style, summary_data)
    
    def print_ratio(self, name, num, denom):
        
        try:
            ratio = num/denom
        except TypeError:
            ratio = "NaN"
            
        print(tabulate([(name, ratio)], tablefmt=self.table_fmt))
        
        return ratio
    
    def print_grouper(self, name, show_group = True, **kwargs):
        
        print(tabulate([(name, "")], tablefmt=self.table_fmt))
        
        if show_group:
            self._group_summarizer(**kwargs)    

    def print_generic(self, *args):

        for table in args:
            
            print(tabulate(table, tablefmt=self.table_fmt))
            
    def print_df(self, data):
        
        print(tabulate(data, headers = data.columns, tablefmt=self.table_fmt))
            
         
            
    
            
        
        
        
        
        


        
        







class NotPandasDataframe(Exception):
    """Raise when data isn't a pandas dataframe"""
    pass

class MissingData(Exception):
    """Raise when data isn't specified"""
    pass

class MissingColumn(Exception):
    """Raise when a column is not provided when necessary"""
    pass
    
class MissingOptimizationMethod(Exception):
    """Raise when an optimization method is not defined for the fit."""
    pass
    
class IncorrectKeys(Exception):
    """ Raise when dictionary keys are incorrectly specified."""
    pass

class NoVars(Exception):
    """Raise when variables not included in the model"""
    pass

class NoConstraints(Exception):
    """Raise when no constraints added to the model"""
    pass

class NoObjective(Exception):
    """Raise when no objective function is in the model"""
    pass

class OptimizationNotInitialized(Exception):
    """Riases when fitter hasn't been called yet."""
    pass


class NonUniqueNutrient(Exception):
    
    def __init__(self, message = "Nutrient occurs multiple times for given vehicle."):
        
        self.message = message
        super().__init__(self.message)

class NutrientNotFound(Exception):
    
    def __init__(self, nutrient,  message=None):
        
        self.nutrient = nutrient
        
        if message is None:
            self.message = f"No nutrient with name {nutrient} found in vehicle_dict"
        super().__init__(self.message)
        
class FunctionValueIncompatible(Exception):
    
    def __init__(self, col, message=None):
        
        self.col = col
        
        if message is None:
            self.message = f"The number of values given to {self.col} does not equal the number of functions"
        
        super().__init__(self.message)
        
class EmptyDataFrame(Exception):
    
    def __init__(self, frame, vehicle, message=None):
        
        self.vehicle = vehicle
        self.frame = frame
        
        if message is None:
            self.message = f"Searching in {self.frame} didn't find {self.vehicle}"


from minimod_opt.solvers.costsolver import CostSolver
from minimod_opt.solvers.benefitsolver import BenefitSolver

class Minimod:
    
    def __new__(self, solver_type = None):
        
        if solver_type == 'covmax':
            
            return BenefitSolver
        
        elif solver_type == 'costmin':
            
            return CostSolver
        


# %%
from typing import Callable
from minimod_opt.continuous.benefits import Benefits
from minimod_opt.continuous.costs import PremixCostCalculator
import pandas as pd
from minimod_opt.continuous.data import CostDataProcessor
import re
import numpy as np
from minimod_opt.continuous.exceptions import NonUniqueNutrient, NutrientNotFound, FunctionValueIncompatible, EmptyDataFrame
from scipy.optimize import minimize, basinhopping

class ContinuousOptimizer:
    def __init__(
        self,
        vehicle_dict: dict,
        benefit_data: pd.DataFrame,
        cost_data: pd.DataFrame,
        benefit_col: str = None,
        increment_col: str = None,
        vehicle_col: str = None,
        nutrient_col: str = None,
        fortificant_col: str = None,
        activity_col: str = None,
        fort_level_col: str = None,
        overage_col: str = None,
        price_col: str = None,
        upcharge: int = 1,
        excipient_price: float = 1.5,
        benefit_func: Callable = None,
        cost_func: Callable = None,
        strict=False,
    ):

        self.vehicle_dict = vehicle_dict
        self.cost_data = cost_data
        self.benefit_data = benefit_data

        self.benefit_col = benefit_col
        self.increment_col = increment_col

        self.vehicle_col = vehicle_col
        self.nutrient_col = nutrient_col
        self.fortificant_col = fortificant_col
        self.activity_col = activity_col
        self.fort_level_col = fort_level_col
        self.overage_col = overage_col
        self.price_col = price_col
        self.upcharge = upcharge
        self.excipient_price = excipient_price

        self.benefit_func = benefit_func
        self.cost_func = cost_func

        self.strict = strict

    def costs(self, nutrient_choice=None):

        cost_dict = {"vehicle": [], "fortificant": [], "cost_fit": [], 'popt_cost' : [], 'pcov_cost' : []}

        for v, f in self.vehicle_dict.items():

            cd = CostDataProcessor(
                data=self.cost_data,
                vehicle=v,
                fortificant=f,
                vehicle_col=self.vehicle_col,
                nutrient_col=self.nutrient_col,
                fortificant_col=self.fortificant_col,
                activity_col=self.activity_col,
                fort_level_col=self.fort_level_col,
                overage_col=self.overage_col,
                price_col=self.price_col
            )

            costs = PremixCostCalculator(
                data=cd, upcharge=self.upcharge, excipient_price=self.excipient_price
            )

            if nutrient_choice is None:
                nutrient, compound = f[0]
            elif isinstance(nutrient_choice, dict):
                nutrient_choice_new = nutrient_choice[v]

                # Use numpy.where to get index of match with true
                compound_index = np.where(
                    [
                        (re.search(fr"{nutrient_choice_new}", x[0], re.IGNORECASE))
                        for x in f
                    ]
                )

                # Now check if len(index) is 1 (it should be, as there should only be one entry for a nutrient in a vehicle)
                if not len(compound_index) == 1:
                    raise NonUniqueNutrient
                elif len(compound_index) == 0:
                    raise NutrientNotFound(nutrient_choice)

                nutrient, compound = f[compound_index[0][0]]
            else:
                raise Exception("Please specify a dictionary")
            
            fit = costs.line_fit(nutrient, compound, self.cost_func)

            cost_dict["vehicle"].append(v)
            cost_dict["fortificant"].append(str(f))
            cost_dict["cost_fit"].append(
                fit[0]
            )
            
            cost_dict['popt_cost'].append(fit[1])
            cost_dict['pcov_cost'].append(fit[2])

        return pd.DataFrame(cost_dict).set_index(["vehicle", "fortificant"])

    @property
    def benefits(self):

        benefit_dict = {"vehicle": [], "fortificant": [], "benefit_fit": [], 'popt_benefit' : [], 'pcov_benefit' : []}

        for v, f in self.vehicle_dict.items():

            df = self.benefit_data.loc[lambda df: df[self.vehicle_col] == v]

            if df.empty:
                raise EmptyDataFrame('benefit data', v)
            
            benefits = Benefits(df, self.benefit_col, self.increment_col)

            fit = benefits.curve_fit(self.benefit_func)
            
            benefit_dict["vehicle"].append(v)
            benefit_dict["fortificant"].append(str(f))
            benefit_dict["benefit_fit"].append(fit[0])
            benefit_dict['popt_benefit'].append(fit[1])
            benefit_dict['pcov_benefit'].append(fit[2])

        return pd.DataFrame(benefit_dict).set_index(["vehicle", "fortificant"])
    
    @property
    def func_len(self):
        return len(self.vehicle_dict)

    def data(self, nutrient):

        df = (
            self.costs(nutrient_choice=nutrient)
            .merge(
                self.benefits,
                left_index=True,
                right_index=True,
                how="outer",
                indicator=True,
            )
        )

        if not (df._merge == "both").all():
            print("[Note]: Merge was not perfect. check `_merge` column")
            if self.strict:
                raise Exception
        else:
            df = df.drop(["_merge"], axis=1)

        return df

    def _dataframe_to_func_sum(self, x, col, nutrient):
        
        funcs = self.data(nutrient = nutrient)[col].values
        
        apply_vectorized = np.vectorize(lambda f, x: f(x), otypes=[object])
        
        if len(x) != len(funcs):
            raise FunctionValueIncompatible(col)
    
        return apply_vectorized(funcs, x).sum()

    def optimize(self, nutrient, benefit_const, x0=None, basin=False, disp=True):
        
        benefit_const_func = lambda x, nutrient: benefit_const - self._dataframe_to_func_sum(x, 'benefit_fit', nutrient)
        
        bounds= [(0, None)]*self.func_len
        constraints = (
            {'type' : 'ineq', 'fun' : benefit_const_func, 'args' : (nutrient, ) }
        )
        
        if basin:
            res = basinhopping(
                self._dataframe_to_func_sum,
                x0,
                minimizer_kwargs= {
                    'method' : 'SLSQP',
                    'args' : ('cost_fit', nutrient),
                    'bounds' : bounds,
                    'constraints' : constraints
                },
                disp=disp
            )
        else:
            res = minimize(self._dataframe_to_func_sum, 
                        x0, 
                        method='SLSQP', 
                        args = ('cost_fit', nutrient),
                        bounds=bounds,
                        constraints=constraints,
                        options = {'disp' : disp})
        
        return res
        

# %%
if __name__ == "__main__":

    from example.example_data import df

    cdf = pd.read_csv("example/cost_data.csv")
    
    ## Trying with real data

    df_real = pd.read_csv("/home/lordflaron/Documents/minimod/examples/data/processed/example1.csv")

    df_real = (
        df_real
        .loc[lambda df: df['space'] == 'Cities']
        .loc[lambda df: df['time']==4]
        .loc[lambda df: df['intervention'].isin(['cube', 'maize', 'oil'])]
        .replace({'intervention' : {'oil' : 'Refined Oil',
                                    'cube' : 'Bouillon',
                                    'maize' : 'Maize Flour'}})
        .rename({'intervention' : 'vehicle'}, axis=1)
    )
    increment= np.linspace(0,500,100)

    func_benefits = lambda x, L, k, x0: L/(1 + np.exp(-k*(x - x0)))

    np.random.seed(1729)
    
    df_incremented = pd.DataFrame()
    
    for i, col in df_real.iterrows():
        
        noise = 0.2 * np.random.normal(size=increment.size)
        benefits = func_benefits(increment, col['benefit'], .3, 1) + noise
        
        df = pd.DataFrame({'increment' : increment,
                           'benefit' : benefits,
                           'vehicle' : col['vehicle']})
        df_incremented = df_incremented.append(df)

    vehicle_dict = {
        "Bouillon": [
            ("Iron", "Micronized ferric pyrophosphate"),
            ("Vitamin A", "Retinyl Palmitate- 250,000 IU/g (dry)"),
            ("Zinc", "Zinc Oxide"),
            ("Vitamin B12", "Vit. B-12 0.1% WS"),
            ("Folic Acid", "Folic Acid"),
        ],
        "Maize Flour": [("Iron", "Ferrous Fumarate"),
                        ('Folic Acid', 'Folic Acid')],
        'Refined Oil' : [('Vitamin A', 
                            "Retinyl Palmitate- 1.7 m IU/g (oil)")]
        
    }

    c = ContinuousOptimizer(
        vehicle_dict=vehicle_dict,
        benefit_data=df_incremented,
        cost_data=cdf,
        benefit_col="benefit",
        increment_col="increment",
        vehicle_col="vehicle",
        fortificant_col="compound",
        activity_col="fort_prop",
        overage_col="fort_over",
    )
    
    nutrient_choice = {"Bouillon": "Vitamin A", 
                      "Maize Flour": "iron",
                      'Refined Oil' : 'Vitamin A'}

    res = c.optimize(nutrient_choice, 80000, [1,1,1])
# %%




from minimod_opt.base.basesolver import BaseSolver
from minimod_opt.utils.exceptions import NotPandasDataframe, MissingColumn
from minimod_opt.utils.summary import OptimizationSummary

from minimod_opt.base.bau_constraint import BAUConstraintCreator
from minimod_opt.base.model import Model


import pandas as pd
import mip
import numpy as np


class CostSolver(BaseSolver):  
    def __init__(self,         
                 all_time=None, 
                 all_space=None, 
                 time_subset = None, 
                 space_subset = None,
                 strict = False,
                 main_constraint_over = None, 
                 minimum_benefit =None,
                 drop_bau = False,
                 **kwargs):

        super().__init__(sense=mip.MINIMIZE, **kwargs)
        
        self.bau = BAUConstraintCreator()

        if isinstance(minimum_benefit, float) or isinstance(minimum_benefit, int):
            self.minimum_benefit = minimum_benefit
            over=None
        elif isinstance(minimum_benefit, str):
            # Get sum of benefits for interventions
            if main_constraint_over == 'space':
                over = self.space_col
            elif main_constraint_over == 'time':
                over = self.time_col
            elif main_constraint_over == 'both':
                over = [self.space_col, self.time_col]
            elif main_constraint_over is None:
                over = None
            self.minimum_benefit = self.bau.create_bau_constraint(
                self._df, minimum_benefit, "discounted_benefits", over
            )
        
            self.bau_df = self.bau.bau_df(
                self._df,
                minimum_benefit,
                [
                    self.benefit_col,
                    self.cost_col,
                    "discounted_benefits",
                    "discounted_costs",
                ],
            )
            
            summed_bau_df = self.bau_df.sum()
            
            self.cost_per_benefit = summed_bau_df[self.cost_col]/summed_bau_df[self.benefit_col]
            
            if drop_bau:
                self._df = self._df.drop(minimum_benefit, level=self.intervention_col)
                   
        self.model = Model(data = self._df, sense = self.sense, solver_name=self.solver_name, show_output=self.show_output)
        
        ## Create base model
        self.model.base_model_create(self.intervention_col, 
                                    self.space_col, 
                                    self.time_col, 
                                    all_time=all_time, 
                                    all_space=all_space, 
                                    time_subset = time_subset, 
                                    space_subset = space_subset,
                                    strict = strict)

        # Add objective and constraint
        self.model.add_objective(self._objective())
        self.model.add_constraint(self._constraint(over = over), self.minimum_benefit, name = "base_constraint")
        

    def _objective(self):

        cost = "discounted_costs"

        # Discounted costs
        return self._discounted_sum_all(col_name = cost)

    def _constraint(self, over = None):

        benefit = "discounted_benefits"
        
        if over is not None:
            return self._discounted_sum_over(benefit, over = over)

        ## Make benefits constraint be at least as large as the one from the minimum benefit intervention
        return self._discounted_sum_all(benefit)

    def fit(self, sol_num=None, **kwargs):
        return self._fit(**kwargs)
    
    def report(self, sol_num = None, intervention_groups = False, quiet=False):

        s = OptimizationSummary(self)

        super().report(sol_num=sol_num, quiet=quiet)
        
        if quiet:
            return
        if self.num_solutions == 1:
            obj_values = self.objective_value
        elif self.num_solutions > 1:
            obj_values = self.objective_values

        sum_costs = self.opt_df["opt_costs_discounted"].sum()
        sum_benefits = self.opt_df["opt_benefit_discounted"].sum()

        results = [
            ("Minimum Benefit", self.minimum_benefit),
            ("Objective Bounds", obj_values),
            ("Total Cost", sum_costs),
            ("Total " + self.benefit_title, sum_benefits),
        ]

        s.print_generic(results)
        s.print_ratio(name="Cost per Benefit", num=sum_costs, denom=sum_benefits)

        s.print_grouper(
            name="Total Cost and Benefits over Time",
            data=self.opt_df[["opt_vals", "opt_benefit", "opt_costs"]],
            style="markdown",
        )
        
        print()
        print("Optimal Interventions")
        print()
        opt_chosen = self.opt_df.loc[lambda df: df['opt_vals']>0]['opt_vals'].unstack(level=self.time_col).fillna(0)
        
        # s.print_df(opt_chosen)
        
        if isinstance(intervention_groups, dict):
            opt_chosen_reset = opt_chosen.reset_index()
            for intervention in intervention_groups:
                opt_chosen_reset[intervention_groups[intervention]] \
                    = opt_chosen_reset[self.intervention_col].str.contains(intervention).astype(int)
            
            intervention_grouper_df = (
                opt_chosen_reset
                .drop(columns = [self.intervention_col, 'opt_vals'])
                .set_index([self.space_col, self.time_col])
                .stack()
                .unstack(level=self.time_col)
                )

            s.print_df(intervention_grouper_df)
                
            # Use for later
            # a = intervention_grouper_df.reset_index(level='region')[list(range(1,11))].mul(intervention_grouper_df.reset_index(level='region')['region'], axis='index')
            # b = a.groupby(a.index).agg(list)

        return opt_chosen
        
        
        




# Imports for local packages
from typing import Union, Type, List, Iterable
from time import time
import matplotlib as mpl

from minimod_opt.utils.exceptions import (
    MissingData,
    NotPandasDataframe,
    MissingOptimizationMethod,
)
from functools import reduce
from pandas import DataFrame

from minimod_opt.utils.summary import OptimizationSummary
from minimod_opt.base.bau_constraint import BAUConstraintCreator
from minimod_opt.utils.plotting import Plotter
from minimod_opt.utils.suppress_messages import suppress_stdout_stderr
from mip.cbc import OptimizationStatus

import matplotlib.pyplot as plt

import pandas as pd
import geopandas as gpd
import numpy as np
import mip
import re

import sys
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.gridspec as gridspec
import matplotlib.ticker as tick
from dataclasses import dataclass
import textwrap
from adjustText import adjust_text



@dataclass
class SupplyCurve:
    
    supply_curve: DataFrame
    full_population: int
    bau: str
    data: DataFrame
    ec_range: Iterable



from typing import Any
from typing import Union

class BaseSolver:

    
    def __init__(
        self,
        data: pd.DataFrame,
        benefit_col: str = 'benefits',
        cost_col: str = 'costs',
        intervention_col: str = 'intervention',
        space_col: str = 'space',
        time_col: str = 'time',
        interest_rate_cost: float = 0.0,  # Discount factor for costs
        interest_rate_benefit: float = 0.03,  # Discount factor for benefits 
        va_weight: float = 1.0,  # VA Weight
        sense: str = None,  # MIP optimization type (maximization or minimization)
        solver_name: str = mip.CBC,  # Solver for MIP to use
        show_output: bool = True,
        benefit_title: str = "Benefits",
    ):       
      
        """The base solver for the Optimization. This sets up the basic setup of the model, which includes:
        - data handling
        - BAU constraint creation
        - base model constraint creation

        Args:
                data (pd.DataFrame): dataframe with benefits and cost data.
                benefit_col (str, optional): name of dataframe's column with benefit data. Defaults to 'benefits'.
                cost_col (str, optional): name of dataframe's column with cost data. Defaults to 'costs'.
                intervention_col (str, optional): name of dataframe's column with intervention data. Defaults to 'intervention'.
                space_col (str, optional): name of dataframe's column with space/region data . Defaults to 'space'.
                time_col (str, optional): name of dataframe's column with time period data . Defaults to 'time'.
                interest_rate_cost (float, optional): interest rate of costs. Defaults to 0.0.
                benefit_title (str, optional): title for benefits to use in plots and reports. Defaults to "Benefits".
        
        ``BaseSolver`` is inherited by ``CostSolver`` and ``BenefitSolver`` to then run optimizations.
        """        
                
        self.interest_rate_cost = interest_rate_cost
        self.interest_rate_benefit = interest_rate_benefit
        self.va_weight = va_weight
        self.discount_costs = 1 / (1 + self.interest_rate_cost)
        self.discount_benefits = 1 / (1 + self.interest_rate_benefit)
        self.benefit_title = benefit_title
        
        if sense is not None:
            self.sense = sense
        else:
            raise Exception("No Optimization Method was specified")
        
        self.solver_name = solver_name
        self.show_output = show_output
        
        # Process Data
        
        if self.show_output:
            print("[Note]: Processing Data...")

        self._df = self._process_data(
            data=data,
            intervention=intervention_col,
            space=space_col,
            time=time_col,
            benefits=benefit_col,
            costs=cost_col,
        )
        
        
        self.benefit_col = benefit_col
        self.cost_col = cost_col
        self.intervention_col = intervention_col
        self.space_col = space_col
        self.time_col = time_col
        
        if self.show_output:
            print("[Note]: Creating Base Model with constraints")
            
        
        self.minimum_benefit = None
        self.total_funds = None
        
        self.message = f"""
                MiniMod Nutrition Intervention Tool
                Optimization Method: {str(self.sense)}
                Solver: {str(self.solver_name)},
                Show Output: {self.show_output}
                
                """
        
        if self.show_output:
            print(self.message)
        
    

    def _discounted_sum_all(self, col_name:str) -> mip.LinExpr:
       
        """Multiply each ``mip_var`` in the data by benefits or costs (``data``) and then create a ``mip`` expression from it.

        Args:
            col_name (str): name of dataframe's column with benefits or costs data

        Returns:
            mip.LinExpr: ``mip`` Expression
        """    
        
        eq = (self._df['mip_vars'] * self._df[col_name]).agg(mip.xsum)

        return eq

    def _discounted_sum_over(self, col_name: str, over: str) -> pd.DataFrame :
        """Abstract function used for constructing the objective function and main constraint of the model

        Args:
            col_name (str): name of dataframe's column with benefits or costs data
            over (str): name of dataframe's column with attribute used to group data by (e.g. time)

        Returns:
            (pd.Dataframe): pd.Dataframe with mip variables as observations
        """
        
        # Merge data with self._df   
        eq = (self._df['mip_vars'] * self._df[col_name]).groupby(over).agg(mip.xsum)
        
        return eq.to_frame().rename({0 : col_name + '_vars'}, axis=1)

 
    def _is_dataframe(self, data: Any):
        """Checks if input dataset is a ``pandas.DataFrame``

        Args:
            data (Any): input data

        Raises:
            NotPandasDataframe: Exception if not a ``pandas.DataFrame``
        """    

        if not isinstance(data, pd.DataFrame):
            raise NotPandasDataframe(
                "[Error]: Input data is not a dataframe. Please input a dataframe."
            )


    def _process_data(
        self,
        data: pd.DataFrame = None,
        intervention: str ="intervention",
        space: str = "space",
        time: str = "time",
        benefits: str = "benefits",
        costs: str = "costs",
    ) -> pd.DataFrame :      
        """Processes the input data by creating discounted benefits and costs.

        Args:
            data (pd.DataFrame, optional): raw data to be processed. Defaults to None.
            intervention (str, optional): name of dataframe's column with intervention data. Defaults to "intervention".
            space (str, optional): name of dataframe's column with space/region data. Defaults to "space".
            time (str, optional): name of dataframe's column with time period data. Defaults to "time".
            benefits (str, optional): name of dataframe's column with benefits data. Defaults to "benefits".
            costs (str, optional): name of dataframe's column with cost data. Defaults to "costs".
        
        Returns:
            (pd.DataFrame): dataframe ready to be used in the problem

        
        |k     | j   |t   | benefits   | costs |
        |------|-----|----|------------|-------|
        |maize |north|0   | 100        | 10    |
        |maize |south|0   | 50         | 20    |
        |maize |east |0   | 30         |30     |
        |maize |west |0   | 20         |40     |
        
        """

        ## First do some sanity checks
        if data is None:
            raise MissingData("No data specified.")

        # Check if dataframe
        self._is_dataframe(data)

        df_aux = (
            data.reset_index()
            .assign(
                time_col=lambda df: df[time].astype(int),
                time_rank=lambda df: (
                    df[time].rank(numeric_only=True, method="dense") - 1
                ).astype(int),
                time_discount_costs=lambda df: self.discount_costs ** df["time_rank"],
                time_discount_benefits=lambda df: self.discount_benefits
                ** df["time_rank"],
                discounted_costs=lambda df: df["time_discount_costs"] * df[costs],
                discounted_benefits=lambda df: df["time_discount_benefits"]*df[benefits]
            )
        )
        
        df_aux[intervention] = df_aux[intervention].str.lstrip().str.rstrip()
        
        df = (
            df_aux
            .set_index([intervention, space, time])
            .sort_index(level=(intervention, space, time))
        )

        return df

    def _constraint(self):
        """This function defines the constraints for the mips model.
        To be overridden by BenefitSolver and CostSolver classes.
        """
        self.constraint_called = 0

    def _objective(self):
        """This function defines the objective function for a model.
        To be overridden by BenefitSolver and CostSolver classes.
        """
        pass
        

    def _fit(self, **kwargs) -> str:
        """Fits data to model. The instantiation of the class creates the base model. Uses ``mip.optimize`` to find the optimal point.

        Args:
            *kwargs: Other parameters to send to mip.optimize

        Returns:
            str: return self
        """
      

        if self.show_output:
            print("[Note]: Optimizing...")

        self.model.optimize(**kwargs)  
        
        (self.objective_value,
         self.objective_values, 
         self.objective_bound, 
         self.num_solutions, 
         self.num_cols, 
         self.num_rows, 
         self.num_int, 
         self.num_nz, 
         self.status) = self.model.get_model_results()
        
        return self
        
    def write(self, filename:str="model.lp"):
        """Save model to file

        Args:
            filename (str, optional):name of the file. Defaults to "model.lp".
        """
        
        self.model.write(filename)
        
    def process_results(self, sol_num:int=None):
        """Processes results of optimization to be used in visualization and reporting functions

        Args:
            sol_num (int, optional): index of solution. Defaults to None.
        """
        
        self.opt_df = self.model.process_results(self.benefit_col, 
                                            self.cost_col, 
                                            self.intervention_col,
                                            self.space_col,
                                            sol_num=sol_num)


    def report(self, sol_num:int=None, quiet:bool=False) -> str:
        """Prints out a report of optimal model parameters and useful statistics.

        Args:
            sol_num (int, optional): index of solution to be displayed. Defaults to None.
            quiet (bool, optional): whether we want the report printed out or not. Defaults to False.
        """
             
        self.opt_df = self.model.process_results(self.benefit_col, 
                                            self.cost_col, 
                                            self.intervention_col,
                                            self.space_col,
                                            sol_num=sol_num)

        if quiet:
            return
        header = [
            ('MiniMod Solver Results', ""),
            ("Method:" , str(self.sense)),
            ("Solver:", str(self.solver_name)),
            ("Optimization Status:", str(self.status)),
            ("Number of Solutions Found:", str(self.num_solutions))

        ]
        
        features = [
            ("No. of Variables:", str(self.num_cols)),
            ("No. of Integer Variables:", str(self.num_int)),
            ("No. of Constraints", str(self.num_rows)),
            ("No. of Non-zeros in Constr.", str(self.num_nz))
        ]
        
        s = OptimizationSummary(self)

        s.print_generic(header, features)
        
        print("Interventions Chosen:")
        
    @property
    def optimal_interventions(self) -> list:
        opt_intervention = (
            self.opt_df
            .loc[lambda df: df['opt_vals']>0]
            .index
            .get_level_values(level=self.intervention_col)
            .unique()
            .tolist()
        )
        """Outputs the unique set of optimal interventions as a list

        Returns:
            list: The list of optimal interventions
        """
        
        return opt_intervention
    
    @property
    def _intervention_list_space_time(self) -> pd.DataFrame:
        """Returns a data frame with multindex (space_col, time_col) where each row is the optimal intervention.

        Returns:
            pd.DataFrame: A dataframe where each row is the optimal intervention for each time period and space
        """
        
        df = (
            self.opt_df['opt_vals']
            .reset_index(level=self.intervention_col)
            .assign(int_appeared= lambda df: df[self.intervention_col]*df['opt_vals'].astype(int))
            .groupby([self.space_col, self.time_col])
            ['int_appeared']
            .agg(set)
            .str.join('')
        )
        
        return df
    
    @property
    def bau_list(self) -> pd.DataFrame:
        """Returns a dataframe with the name of the bau intervention. Mostly done for compatibility with other methods.

        Returns:
            pd.DataFrame: dataframe with the name of the bau intervention
        """
        
        df = (
            self.bau_df
            .reset_index(level=self.intervention_col)
            .rename({self.intervention_col : 'int_appeared'}, axis=1)
            ['int_appeared']
        )
        
        return df
     
        
    def plot_time(self, 
                  fig: mpl.figure = None, 
                  ax: mpl.axis= None,
                  save: str = None,
                  cumulative: bool = False,
                  cumulative_discount: bool = False) -> mpl.figure:
        """Plots optimal benefits and costs across time after model optimization

        Args:
            fig (matplotlib.figure, optional): matplotlib figure. Defaults to None.
            ax (matplotlib.axis, optional): matplotlib axis to use. Defaults to None.
            save (str, optional): path to save the figure. Defaults to None.
            cumulative (bool, optional): whether to plot cumulative benefits or costs. Defaults to False.
            cumulative_discount (bool, optional): whether to plot cumulative benefits or costs, discounted. Defaults to False.

        Returns:
            matplotlib.figure: figure with optimal benefits and cost across time
        """
        
        p = Plotter(self)
                
        if cumulative:
            return p._plot_lines(to_plot = ['cumulative_benefits', 'cumulative_costs'],
                    title= "Optima over Time",
                    xlabel = 'Time',
                    ylabel = self.benefit_title,
                    twin =True,
                    twin_ylabel= "Currency",
                    save = save,
                    legend = ['Cumm. ' + self.benefit_title,
                            'Cumm. Costs'],
                    figure=fig,
                    axis=ax)
        elif cumulative_discount:
            return p._plot_lines(to_plot = ['cumulative_discounted_benefits', 'cumulative_discounted_costs'],
                    title= "Optima over Time",
                    xlabel = 'Time',
                    ylabel = self.benefit_title,
                    twin =True,
                    twin_ylabel= "Currency",
                    save = save,
                    legend = ['Cumm. Dis. '+ self.benefit_title,
                            'Cumm. Dis. Costs'],
                    figure=fig,
                    axis=ax)
        else:
            return p._plot_lines(to_plot = ['opt_benefit', 'opt_costs'],
                                title= "Optima over Time",
                                xlabel = 'Time',
                                ylabel = self.benefit_title,
                                twin =True,
                                twin_ylabel= "Currency",
                                save = save,
                                legend = ['Optimal ' + self.benefit_title,
                                        'Optimal Costs'],
                                figure=fig,
                                axis=ax)

    def plot_bau_time(self,
                      opt_variable: str = 'b',
                      fig: mpl.figure = None,
                      ax: mpl.axis = None,
                      save: str = None):
        """Plots benefits and costs of optimal and benchark interventions across time 

        Args:
            opt_variable (str, optional): optimal variable to be plotted, where
                b = optimal benefits
                c = 'optimal costs
                cdb = cumulative discounted benefits
                cdc = cumulative discounted costs
                cb = cumulative benefits
                cc = cumulative costs
            Defaults to 'b'.
            fig (matplotlib.figure, optional): matplotlib figure. Defaults to None.
            ax (matplotlib.axis, optional):matplotlib axis to use. Defaults to None.
            save (str, optional): path to save the figure. Defaults to None.

        Raises:
            Exception: not one of the allowed variables for map plotting
        """
        
        if ax is None:
            fig, ax = plt.subplots()

        p = Plotter(self)
        
        if opt_variable == 'b':
            opt = 'opt_benefit'
            bau_col = self.benefit_col
            title = "Optimal " + self.benefit_title + " vs. BAU"
        elif opt_variable == 'c':
            opt = 'opt_costs'
            bau_col = self.cost_col
            title = "Optimal Costs vs. BAU"
        elif opt_variable == 'cdb':
            opt = 'cumulative_discounted_benefits'
            bau_col = 'discounted_benefits'
            title = 'Cumulative Discounted ' + self.benefit_title
        elif opt_variable == 'cdc':
            opt = 'cumulative_discounted_costs'
            bau_col = 'discounted_costs'
            title = 'Cumulative Discounted Costs'
        elif opt_variable == 'cb':
            opt = 'cumulative_benefits'
            bau_col = self.benefit_col
            title = 'Cumulative ' + self.benefit_title      
        elif opt_variable == 'cc':
            opt = 'cumulative_costs'
            bau_col = self.cost_col
            title = 'Cumulative Costs'
        else:
            raise Exception("Not one of the allowed variables for map plotting. Try again.")
        
        if opt_variable in ['cdb', 'cdc', 'cb', 'cc']:
            
            bench_df = (
                self.bau_df
                .groupby([self.time_col])
                .sum().cumsum()
                .assign(bench_col = lambda df: df[bau_col])
                )
            
        else:
            bench_df = (
                self.bau_df
                .assign(bench_col = lambda df: df[bau_col])
                .groupby([self.time_col])
                .sum()
                        )

        p._plot_lines(to_plot = opt,
                    title= title,
                    xlabel = 'Time',
                    figure=fig,
                    axis=ax)
        
        bench_df['bench_col'].plot(ax=ax)
        
        ax.legend(['Optimal', 'BAU'])
        ax.set_xlabel('Time')

        if save is not None:
            plt.savefig(save)

             

    def plot_opt_val_hist(self, 
                          fig: mpl.figure = None, 
                          ax: mpl.axis = None, 
                          save: str = None) -> mpl.figure:
        """A histogram of the optimally chosen interventions

        Args:
            fig (matplotlib.figure, optional): figure instance to use. Defaults to None.
            ax (matplotlib.axis, optional): axis instance to use. Defaults to None.
            save (str, optional): path to save the figure. Defaults to None.

        Returns:
            matplotlib.figure: histogram figure
        """      
        
        p = Plotter(self)
        
        return p._plot_hist(to_plot = 'opt_vals',
                            title = "Optimal Choices",
                            xlabel = "Time",
                            ylabel= "",
                            figure = fig,
                            axis = ax,
                            save = save)  

    def plot_chloropleth(self,
                         intervention: str = None,
                         time: Union[int,list] = None,
                         optimum_interest: str = 'b',
                         map_df: gpd.GeoDataFrame  = None,
                         merge_key: Union[str,list] = None,
                         intervention_bubbles: bool = False,
                         intervention_bubble_names: Union[str,list] = None,
                         save: str = None):
        """Creates a chloropleth map of the specified intervention and time period for the optimal variable. 
        If more than one intervention is specified, then aggregates them. If more than one time period is specified, then creates a subplots of len(time) and show each.

        Args:
            intervention (str, optional): intervention to use. Defaults to None.
            time (Union[int,list], optional): time periods to plot. Defaults to None.
            optimum_interest (str, optional): optimal variable to use (Options include: 'b' for optimal benefits, 'c' for optimal costs, and 'v' for optimal variable). Defaults to 'b'.
            map_df (geopandas.GeoDataFrame, optional): geopandas dataframe with geometry information. Defaults to None.
            merge_key (Union[str,list], optional): column to merge geo-dataframe. Defaults to None.
            intervention_bubbles (bool, optional): whether to show optimal intervention names in map. Defaults to False.
            intervention_bubble_names (Union[str,list], optional): key to merge on to geo dataframe. Defaults to None.
            save (str, optional): path to save map. Defaults to None.

        Raises:
            Exception: Not one of the allowed variables for map plotting
        """
        
        
        if intervention is None:
            intervention = self.optimal_interventions
        
        p = Plotter(self)
                
        if optimum_interest == 'b':
            opt = 'opt_benefit'
            title = self.benefit_title
        elif optimum_interest == 'c':
            opt = 'opt_costs'
            title = "Optimal Costs"
        elif optimum_interest == 'v':
            opt = 'opt_vals'
            title = "Optimal Interventions"
        elif optimum_interest == 'cdb':
            opt = 'cumulative_discounted_benefits'
            title = 'Cumulative Discounted ' + self.benefit_title
        elif optimum_interest == 'cdc':
            opt = 'cumulative_discounted_costs'
            title = 'Cumulative Discounted Costs'
        elif optimum_interest == 'cb':
            opt = 'cumulative_benefits'
            title = 'Cumulative ' + self.benefit_title      
        elif optimum_interest == 'cc':
            opt = 'cumulative_costs'
            title = 'Cumulative Costs'
        else:
            raise Exception("Not one of the allowed variables for map plotting. Try again.")
        
        
        plotter = p._plot_chloropleth_getter(time = time)
        plot = plotter(data = self.opt_df,
                                          intervention = intervention,
                                            time = time,
                                            optimum_interest=opt,
                                            map_df = map_df,
                                            merge_key=merge_key,
                                            aggfunc = 'sum',
                                            title = title,
                                            intervention_bubbles = intervention_bubbles,
                                            intervention_bubble_names = intervention_bubble_names,
                                            save = save)
        # return plot


    def plot_grouped_interventions(self, 
                                    data_of_interest: str = 'benefits', 
                                    title: str = None,
                                    intervention_subset: Union[str,list] = slice(None),
                                    save: str = None):
        """Shows Optimal level of benefits or costs in a grouped bar plots for every optimally chosen variable across regions.

        Args:
            data_of_interest (str, optional): variable to show. Defaults to 'benefits'.
            title (str, optional):title for resulting plot. Defaults to None.
            intervention_subset (Union[str,list], optional): subset of interventions to show on bar plot. Defaults to slice(None).
            save (str, optional): path to save the figure. Defaults to None.
        """
                                 
        
        p = Plotter(self)
            
        if data_of_interest == 'benefits':
            col_of_interest = 'opt_benefit'
        elif data_of_interest == 'costs':
            col_of_interest = 'opt_costs'
        
        p._plot_grouped_bar(intervention_col= self.intervention_col,
                            space_col = self.space_col,
                            col_of_interest= col_of_interest,
                            ylabel = "Optimal Level",
                            intervention_subset= intervention_subset,
                            save = save)
        
     

    def plot_map_benchmark(self,
                           intervention: list = None,
                           time: list = None,
                           optimum_interest: str = 'b',
                           bench_intervention: list = None,
                           map_df: gpd.GeoDataFrame = None,
                           merge_key: Union[str,list] = None,
                           save: str = None,
                           intervention_in_title: bool = True,
                           intervention_bubbles: bool = False,
                           intervention_bubble_names: Union[str,list] = None,
                           millions: bool = True,
                           bau_intervention_bubble_names: Union[str,list] = None
                           ):
        """Maps the optimal level on a map against a benchmark, optionally the BAU level chosen from ``minimum_benefit`` or ``total_funds``.

        Args:
            intervention (list, optional): interventions to map. Defaults to None.
            time (list, optional): time periods to map. Defaults to None.
            optimum_interest (str, optional):  optimal value to use. Options include 'b' (benefits), 'c' (costs), 'v' (variables). Defaults to 'b'.
            bench_intervention (list, optional): interventions to use for benchmark. Defaults to None.
            map_df (geopandas.GeoDataFrame, optional):  geo dataframe with geometry data. Defaults to None.
            merge_key (Union[str,list], optional): key to merge data from opt_df to geo dataframe. Defaults to None.
            save (str, optional): path to save the map. Defaults to None.
            intervention_in_title (bool, optional): True if intervention name will be included in the title of the figure. Defaults to True.
            intervention_bubbles (bool, optional): True if intervention bubbles. Defaults to False.
            intervention_bubble_names (Union[str,list], optional): names of intervention bubbles. Defaults to None.
            millions (bool, optional): True if values displayed in millions. Defaults to True.
            bau_intervention_bubble_names (Union[str,list], optional): name for bau intervention bubble. Defaults to None.

        """
                
        if intervention is None:
            intervention = self.optimal_interventions   
            
        if figsize is not None:   
            fig = plt.figure(figsize=figsize)
        else:
            fig = plt.figure(figsize=(10,12))
        
        gs = gridspec.GridSpec(2,2, height_ratios = [6,1])
        optimal = fig.add_subplot(gs[0,0])
        bench = fig.add_subplot(gs[0,1])
        cbar = fig.add_subplot(gs[1,:])
        
        p = Plotter(self)
        
        if optimum_interest == 'b':
            opt = 'opt_benefit'
            bench_col = self.benefit_col
            title = self.benefit_title
        elif optimum_interest == 'c':
            opt = 'opt_costs'
            bench_col = self.cost_col
            title = "Costs"
        elif optimum_interest == 'v':
            opt = 'opt_vals'
            title = "Interventions"
        elif optimum_interest == 'cdb':
            opt =  'cumulative_discounted_benefits'
            bench_col = 'discounted_benefits'
            title = 'Cumulative Discounted ' + self.benefit_title
        elif optimum_interest == 'cdc':
            opt =  'cumulative_discounted_costs'
            title = 'Cumulative Discounted Costs'
            bench_col = 'discounted_costs'
        elif optimum_interest == 'cb':
            opt =  'cumulative_benefits'
            title = 'Cumulative ' + self.benefit_title   
            bench_col = self.benefit_col   
        elif optimum_interest == 'cc':
            opt =  'cumulative_costs'
            title = 'Cumulative Costs'
            bench_col = self.cost_col
        else:
            raise Exception("Not one of the allowed variables for map plotting. Try again.")
        
        if bench_intervention is None:
            bench_intervention = self.minimum_benefit
        
        if merge_key is None:
            merge_key = self.space_col
            
        if optimum_interest in ['cdb', 'cdc', 'cb', 'cc']:
            
            bench_df = self.bau_df.assign(bench_col = lambda df: (df
                                                            .groupby([self.intervention_col, 
                                                                    self.space_col])
                                                            [bench_col]
                                                            .transform('cumsum')))
            
        else:
            bench_df = self.bau_df.assign(bench_col = lambda df: df[bench_col])
        
        y = 1.05
        
        if intervention_in_title:
            title = title + f"\nOptimal Interventions:\n{', '.join(intervention)}"
            y = y + .05
            
        fig.suptitle(title, y=y)
        plotter = p._plot_chloropleth_getter(time)
        
        # Get min and max values for color map
        opt_max = self.opt_df[opt].max()
        opt_min = self.opt_df[opt].min()
        
        bench_max = bench_df['bench_col'].max()
        bench_min = bench_df['bench_col'].min()
        
        vmax = max(opt_max, bench_max)
        vmin = min(opt_min, bench_min)
        
        if intervention_bubbles:
            bau_intervention_bubbles = 'bau'
        else:
            bau_intervention_bubbles = False
        
        
        plotter(data = self.opt_df,
                    intervention = intervention,
                    time = time,
                    optimum_interest=opt,
                    map_df = map_df,
                    merge_key=merge_key,
                    aggfunc = 'sum',
                    ax = optimal,
                    cax = cbar,
                    title = "Optimal Scenario",
                    intervention_bubbles = intervention_bubbles,
                    intervention_bubble_names = intervention_bubble_names,
                    vmin = vmin,
                    vmax = vmax,
                    legend_kwds = {'orientation' : 'horizontal'})
        
        plotter(data = bench_df,
                        intervention = bench_intervention,
                        time = time,
                        optimum_interest= 'bench_col',
                        map_df = map_df,
                        merge_key=merge_key,
                        aggfunc = 'sum',
                        ax = bench,
                        show_legend = False,
                        title = f"BAU* Scenario",
                         vmin = vmin,
                        vmax = vmax,
                        intervention_bubbles = bau_intervention_bubbles,
                        intervention_bubble_names = bau_intervention_bubble_names)
        
        plt.tight_layout()
        
        fig_text = 'Note: Colors describe ' + title
        
        if millions:
            fig_text = fig_text + ' (in millions)'
        
        # if intervention_bubble_names:
        #     fig_text = fig_text + '\nBAU* scenario made up of ' + ', '.join(intervention_bubble_names)
        
        fig.text(0.5,-.05, fig_text, ha='center')
                
        if save is not None:
            plt.savefig(save, dpi = p.dpi, bbox_inches="tight")
    
    @classmethod  
    def supply_curve(cls, data, 
                    full_population,
                    bau,
                    all_space,
                    all_time,
                    time_subset,
                    benefit_col='effective_coverage',
                    cost_col='cost',
                    intervention_col='intervention',
                    space_col='space',
                    time_col='time',
                    ec_range=None,
                    above_ul = False,
                    **kwargs): 
        
        if ec_range is None:
            ec_range = np.arange(.1,1,.1)
            
        
        ratio_to_constraint = lambda ratio: ratio*full_population
        
        model_dict = {}

        for _, benefit_constraint in enumerate([bau] + list(ec_range)):
            
            if isinstance(benefit_constraint, str):
                minimum_benefit = benefit_constraint
            else:
                minimum_benefit = ratio_to_constraint(benefit_constraint)
            
            c = cls(minimum_benefit = minimum_benefit,
                        data = data, 
                        benefit_col = benefit_col,
                        cost_col = cost_col,
                        intervention_col = intervention_col,
                        space_col = space_col,
                        time_col = time_col,
                        all_space =all_space, 
                        all_time = all_time,
                        time_subset = time_subset,
                        **kwargs)

            opt = c.fit()
            
            model_dict[benefit_constraint] = opt
            
        results_dict = {'benefit' : [],
                        'opt_benefits' : [],
                        'opt_costs' : [],
                        'opt_interventions' : [],
                        'convergence' : [],
                        'vas_regions' : []}
        if above_ul:
            results_dict['opt_above_ul'] = []

        
        for benefit_constraint, model in model_dict.items():

            model.report(quiet=True)
            results_dict['benefit'].append(benefit_constraint)
            results_dict['opt_benefits'].append(model.opt_df.opt_benefit_discounted.sum())
            results_dict['opt_costs'].append(model.opt_df.opt_costs_discounted.sum())
            results_dict['opt_interventions'].append(model.optimal_interventions)
            results_dict['convergence'].append(model.status)
            results_dict['vas_regions'].append(model.opt_df
                                               .query("opt_vals > 0")
                                               .query("index.get_level_values('intervention').str.contains('vas')")
                                               .index.get_level_values('region').unique().values.tolist())
            
            if above_ul:
                above_ul_df = (
                        data['above_ul']
                        .reset_index()
                        .assign(intervention = lambda df: df[intervention_col].str.lower())
                        .set_index([intervention_col, space_col, time_col])
                        )
                
                opt_above_ul = (model.opt_df['opt_vals'] * above_ul_df['above_ul']).sum()
                
                results_dict['opt_above_ul'].append(opt_above_ul)
                
        return SupplyCurve(pd.DataFrame(results_dict, index=[results_dict['opt_benefits'][0]/full_population] + list(ec_range)) 
                           , full_population=full_population,
                           bau=bau,
                           data=data,
                           ec_range=ec_range)        
        
        
    @classmethod      
    def plot_supply_curve(cls, 
                        supply_curve: SupplyCurve = None,
                          data: DataFrame = None, 
                    full_population: Union[int, float] = None,
                    bau: str=None,
                    all_space: List[float]=None,
                    all_time=None,
                    time_subset=None,
                    benefit_col='effective_coverage',
                    cost_col='cost',
                    intervention_col='intervention',
                    space_col='space',
                    time_col='time',
                    ec_range=None,
                    above_ul = False,
                    ec_thousands = 1_000,
                    ul_thousands = 1_000,
                    save=None,
                    splitter=None,
                    **kwargs):
        
        if supply_curve is None:
            supply_curve = cls.supply_curve(
                        data=data, 
                        full_population=full_population,
                        bau=bau,
                        all_space=all_space,
                        all_time=all_time,
                        time_subset=time_subset,
                        benefit_col=benefit_col,
                        cost_col=cost_col,
                        intervention_col=intervention_col,
                        space_col=space_col,
                        time_col=time_col,
                        ec_range=ec_range,
                        above_ul = above_ul,
                        **kwargs
                        )
        
        N_ec = len(supply_curve.ec_range)
        full_population = supply_curve.full_population
        bau = supply_curve.bau
        data=supply_curve.data
        supply_curve = supply_curve.supply_curve # Now that we have all info. make `supply_curve` just the dataframe

        # make nan of infeasible solutions
        supply_curve = supply_curve.replace(0, np.nan)
        
        only_bau_opt = supply_curve.query("benefit == @bau")
        
        # # get places where interventions change
        # transitions = supply_curve[~(supply_curve['opt_interventions'] == supply_curve['opt_interventions'].shift(-1))]
        
        # Now get additions of interventions by converting to set and doing difference
        supply_curve = (
            supply_curve
            .dropna()
            .assign(opt_interventions = lambda df: df['opt_interventions'].apply(lambda x: [i.split(' + ') for i in x]))
            .assign(opt_interventions = lambda df: df['opt_interventions'].apply(lambda x: reduce(lambda y, z: y + z, x)))
            .assign(opt_interventions = lambda df: df['opt_interventions'].apply(lambda x: set(x)))
            .assign(int_transitions_lag = lambda df: df['opt_interventions'].shift(-1))
            .dropna()
            .assign(transitions_plus = lambda df: df.apply(lambda col: col['int_transitions_lag'].difference(col['opt_interventions']), axis=1).shift(1).fillna(df['opt_interventions']))
            .assign(transitions_minus = lambda df: df.apply(lambda col: col['opt_interventions'].difference(col['int_transitions_lag']), axis=1).shift(1).fillna(df['opt_interventions']))
            .query("benefit != @bau")
            .loc[lambda df: df.index >= only_bau_opt.index.values[0]]
            )
        
        if above_ul:
            subplot_col = 2
        else:
            subplot_col = 1
        
        with plt.style.context('seaborn-whitegrid'):
            fig, ax = plt.subplots(subplot_col+1, 1, figsize=(5*subplot_col,12))
            
            
            if above_ul:
                ax0 = ax[0]
                ax1 = ax[1]
                ax_region=ax[2]
            else:
                ax0 = ax
                ax_region=ax[1]
                      
            supply_curve['opt_costs'].plot(ax=ax0)
                        
            ax0.set_xlabel('Effective Coverage (%)')

            ax0.set_xticks(supply_curve.index.tolist())
            
            ax0.set_xticklabels([f"{x*100:.0f}" for x in supply_curve.index.tolist()])
            
            ax0.ticklabel_format(axis='y', useMathText=True)
            
            ax0.yaxis.set_major_formatter(tick.FuncFormatter(lambda y, _: f"{y/ec_thousands:,.0f}"))
            
            ax0.set_title(f'Total Cost (x {ec_thousands:,.0f})')
            
            if bau is not None:
                if data is None:
                    raise Exception("If `bau` is specified, `data` must also be specified")
                
                bau_ec, bau_costs = BAUConstraintCreator().bau_df(data, bau, ['effective_coverage', 'costs']).sum()

                ax0.scatter(bau_ec/full_population, bau_costs, color='tab:green', marker='s')
                ax0.annotate("BAU", (bau_ec/full_population, bau_costs), xytext=(5, 5),  
                            textcoords="offset pixels", color='black')
                        
            opt_b, opt_c = only_bau_opt[['opt_benefits', 'opt_costs']].values.tolist()[0]
            
            ax0.annotate("Optimum", (opt_b/full_population, opt_c), xytext=(5, -10),  
                        textcoords="offset pixels", color='black')
            ax0.scatter(opt_b/full_population, opt_c,  color='tab:orange', marker="s")
            
            if above_ul:
                # Now get above_ul
                supply_curve['opt_above_ul'].plot(ax=ax1, color='tab:red')
                ax1.yaxis.set_major_formatter(tick.FuncFormatter(lambda y, _: f"{y/ul_thousands:,.0f}"))
                ax1.set_title(f'Population Above UL (x {ul_thousands:,.0f})')
                ax1.set_xticks(supply_curve.index.tolist())
                ax1.set_xticklabels([f"{x*100:.0f}" for x in supply_curve.index.tolist()])
                ax1.set_xlabel('Effective Coverage (%)')
            
            def message_writer(x):
                if len(x['transitions_plus']) != 0 and len(x['transitions_minus'])!=0:
                    message= "+ " + ', '.join(x['transitions_plus']) +  " - " + ", ".join(x['transitions_minus'])
                    message = '\n'.join(textwrap.wrap(message, width=50))
                    
                    # ns_needed = len(message)//25
                    
                    # # for i in range(ns_needed):
                    # #     message = message[i*25:i+25] + r'-len\n' + message[25+1+i:]
                    
                    return message
            
            # print(supply_curve)

            
            (
                supply_curve['vas_regions']
                .explode()
                .to_frame()
                .assign(yes=lambda df: \
                    (~df['vas_regions'].isnull())
                    .astype(int))
                .set_index('vas_regions', 
                           append=True)
                .unstack()
                .drop(columns=('yes', pd.NA))
                .plot.bar(stacked=True, legend=False, 
                          ax=ax_region, cmap='tab20')
                )
            
            # ax_region.set_xticks(supply_curve.index.tolist())
            ax_region.set_xticklabels([f"{x*100:.0f}" for x in supply_curve.index.tolist()],
                                      rotation=0)
            ax_region.set_xlabel('Effective Coverage (%)')
            ax_region.set_yticklabels([])
            ax_region.set_title("VAS Regions")
            
            texts = (
                supply_curve
                .assign(message = lambda df: df.apply(lambda x: message_writer(x), axis=1))
                .apply(lambda df: ax0.text(df.name+.01, df['opt_costs'], 
                                                s=df['message'],
                                                color='black', wrap=True), axis=1) 
                .values.tolist()
                )   
            
            adjust_text(texts, arrowprops=dict(arrowstyle='->', color='green'))
            
            figtext = f"Note: BAU refers to a nutritional intervention of {bau.title()}." \
            f" Optimum solution consists of {', '.join(only_bau_opt['opt_interventions'].values[0])}." \
            " Vitamin A Supplementation taking place at various level of effective coverage in: " \
                           f"{', '.join(list(set(np.concatenate(supply_curve['vas_regions'].values))))}"
                        
            txt = fig.text(.05, -.1, s='\n'.join(textwrap.wrap(figtext, width=100)))
                             
            handles, labels = ax_region.get_legend_handles_labels()
            
            new_labels = [i.split(',')[1].replace(')', '').strip() for i in labels]
            plt.legend(handles, new_labels, ncol=4, loc='lower center', 
                       bbox_to_anchor = (.5,-.6),
                       title='Regions')                 
                               
            plt.tight_layout()
            
            if save is not None:
                plt.savefig(save, dpi=300, bbox_inches='tight')
            
        return ax
        
        



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['minimod_opt',
 'minimod_opt.base',
 'minimod_opt.continuous',
 'minimod_opt.monte_carlo',
 'minimod_opt.solvers',
 'minimod_opt.tests',
 'minimod_opt.utils']

package_data = \
{'': ['*'], 'minimod_opt.continuous': ['example/*']}

install_requires = \
['adjustText>=0.7.3,<0.8.0',
 'matplotlib>=3.6.1,<4.0.0',
 'mip==1.9.3',
 'numpy>=1.23.4,<2.0.0',
 'pandas>=1.5.0,<2.0.0',
 'pqdm>=0.2.0,<0.3.0',
 'progressbar33>=2.4,<3.0',
 'python-docx>=0.8.11,<0.9.0',
 'tabulate>=0.9.0,<0.10.0',
 'xlrd>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'minimod-opt',
    'version': '0.1.11',
    'description': 'A mixed-integer program optimization solver for finding optimal nutritional interventions across space and time.',
    'long_description': '# minimod-opt\nA mixed-integer solver that solves optimal intervention allocation across time and space (Port from GAMS)\n\nThis python module is designed as a port from GAMS and uses `mip` to do mixed-integer optimization.\n\nVisit [the website](https://minimod-nutrition.github.io/minimod) for more information!\n',
    'author': 'Aleksandr Michuda',
    'author_email': 'am2497@cornell.edu',
    'maintainer': 'Aleksandr Michuda',
    'maintainer_email': 'am2497@cornell.edu',
    'url': 'https://minimod-nutrition.github.io/minimod-opt/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Callable, Dict, List, Union
from minimod_opt.solvers import Minimod
from minimod_opt.utils.plotting import Plotter
from minimod_opt.utils.summary import OptimizationSummary
from mip import OptimizationStatus

import numpy as np
import pandas as pd
from tqdm import tqdm
from pqdm.processes import pqdm

import matplotlib.pyplot as plt

from functools import reduce
from multiprocessing import Pool
from functools import partial
import matplotlib.ticker as mtick


class MonteCarloMinimod:
    def __init__(
        self,
        solver_type: str=None,
        data: pd.DataFrame=None,
        intervention_col: str=None,
        time_col: str=None,
        space_col:str=None,
        benefit_mean_col: str=None,
        benefit_sd_col: str=None,
        cost_col: str=None,
        cost_uniform_perc: float=None,
        pop_weight_col: str=None,
        **kwargs,
    ):
        """MonteCarloMinimod uses the `Minimod` optimization classes and conducts Monte Carlo simulations of benefits and cost data through distribution assumptions on the data. Currently the simulations assume a normal distribution about benefits and a uniform distribution around costs.
        
        

        Args:
            solver_type (str, optional): Whether to minimize costs or maximize benfits. Currently only the former is implemented. Defaults to None.
            data (pd.DataFrame, optional): The input data. Defaults to None.
            intervention_col (str, optional): the name of the intervention variable. Defaults to None.
            time_col (str, optional): the name of the time variable. Defaults to None.
            space_col (str, optional): the name of region/spatial variable. Defaults to None.
            benefit_mean_col (str, optional): the name of the variable that gives the mean of benefits for the intervention. Usually just the name of the benefits variable. Defaults to None.
            benefit_sd_col (str, optional): The name of the variable for the standard deviation of benefits. Defaults to None.
            cost_col (str, optional): The name of the cost variable. Defaults to None.
            cost_uniform_perc (float, optional): If using the default of a uniform distribution for costs, the amount of the endpoint of the uniform distribution ([cost*(1+cost_uniform_perc), cost*(1- cost_uniform_perc)]). Defaults to None.
            pop_weight_col (str, optional): The name name of the variable that specific population weights. Defaults to None.
            
        Examples:
        
            Below is an example of how you would run the simulations as well as the visualizations.
            
            >>> import os; print(os.getcwd())
            >>> df = (
            ...     pd.read_csv("data/processed/example1.csv")
            ...     .assign(benefit_sd = lambda df: df['benefit']/2,
            ...             costs_sd = lambda df: df['costs']/2)
            ...     )

            >>> cube = ["cube", "vascube", "oilcube", "cubemaize", "vascubemaize", "vasoilcube", "oilcubemaize", "vasoilcubemaize"]
            >>> oil = ["oil", "vasoil", "oilcube", "oilmaize", "vasoilmaize", "vasoilcube", "oilcubemaize", "vasoilcubemaize"]
            >>> maize = ["maize", "vasmaize", "oilmaize", "cubemaize", "vascubemaize", "vasoilmaize", "oilcubemaize", "vasoilcubemaize" ]

            >>> a = mm.MonteCarloMinimod(solver_type = 'costmin', 
            ...                        data = df, 
            ...                        intervention_col='intervention',
            ...                        space_col='space',
            ...                        time_col='time',
            ...                        benefit_mean_col = 'benefit',
            ...                        benefit_sd_col= 'benefit_sd',
            ...                        cost_col='costs')

            >>> def benefit_no_change(seed, benefit_col, data):
            ...    return data[benefit_col]
            
            >>> sim = a.fit_all_samples(N = 100, 
            ...                         all_space=oil, 
            ...                         all_time=cube, 
            ...                         time_subset=[1,2,3], 
            ...                         minimum_benefit='vasoilold', 
            ...                         benefit_callable=benefit_no_change, 
            ...                         benefit_kwargs={'benefit_col' : 'benefit'}
            ...                         )
            
            >>> a.plot_opt_hist(save = "sim_results.png")

            >>> a.report(perc_intervention_appeared=True)

            >>> a.plot_sim_trajectories(save = 'sim_traj.png')
        
        """        
     

        print("""Monte Carlo Simulator""")

        self.solver_type = solver_type

        self.data = data.set_index([intervention_col, space_col, time_col])

        self.intervention_col = intervention_col
        self.space_col = space_col
        self.time_col = time_col

        if pop_weight_col is None:
            self.data = self.data.assign(pop_weight_col=1)
            self.pop_weight_col = "pop_weight_col"
        else:
            self.pop_weight_col = pop_weight_col

        self.benefit_mean_col = benefit_mean_col
        self.benefit_sd_col = benefit_sd_col

        if cost_uniform_perc is None:
            self.cost_uniform_perc = 0.2
        else:
            self.cost_uniform_perc = cost_uniform_perc

        self.cost_col = cost_col


    def _construct_benefit_sample(self, seed: int, data: pd.DataFrame=None, benefit_col: str = 'benefit_random_draw') -> pd.Series:
        """Draw of a sample of benefits assuming normality

        Args:
            seed (int): The random seed
            data (pd.DataFrame, optional): The input data. Defaults to None.
            benefit_col (str, optional): the name of the resulting benefits variable. Defaults to 'benefit_random_draw'.

        Returns:
            pd.Series: A series of benefits
        """        

        random = np.random.default_rng(seed=seed)
        
        df_mean_sd = data[
            [self.benefit_mean_col, self.benefit_sd_col, self.pop_weight_col]
        ]

        df = df_mean_sd.pipe(self._drop_nan_benefits).assign(
            weight_mean=lambda df: df[self.benefit_mean_col] * df[self.pop_weight_col],
            weight_sd=lambda df: df[self.benefit_sd_col] * df[self.pop_weight_col],
            benefit_random_draw=lambda df: random.normal(
                df["weight_mean"], df["weight_sd"]
            ),
        )

        return df[benefit_col]

    def _construct_cost_sample(self, seed: str, data: pd.DataFrame=None, cost_col: str='cost_random_draw') -> pd.Series:
        """Draw a sample costs assuming a uniform distribution

        Args:
            seed (str): The randome
            data (pd.DataFrame, optional): The input data. Defaults to None.
            cost_col (str, optional): The name of the cost variable. Defaults to 'cost_random_draw'.

        Returns:
            pd.Series: The cost draw
        """        

        random= np.random.default_rng(seed=seed)

        df_costs = data[self.cost_col]
        df_costs_low = (1 - self.cost_uniform_perc) * data[self.cost_col]
        df_costs_high = (1 + self.cost_uniform_perc) * data[self.cost_col]

        df = df_costs.to_frame().assign(
            cost_random_draw=random.uniform(df_costs_low, df_costs_high)
        )

        return df[cost_col]

    def _drop_nan_benefits(self, data):

        df = data.dropna(subset=[self.benefit_sd_col])

        return df

    def _merge_samples(self, benefit_callable: Callable, cost_callable: Callable,  
                       cost_kwargs: dict, benefit_kwargs: dict) -> pd.DataFrame:
        """Transforms the cost and benefit data and merges them together.

        Args:
            benefit_callable (Callable): The function for transforming benefits
            cost_callable (Callable): The function for transforming costs
            cost_kwargs (dict): extra arguments for `cost_callable`
            benefit_kwargs (dict): extra arguments for `benefit_callable`

        Returns:
            pd.DataFrame: The merged dataset of benefits and costs
        """        
        

        benefit_sample = benefit_callable(**benefit_kwargs)
        cost_sample = cost_callable(**cost_kwargs)

        return benefit_sample.to_frame().merge(
            cost_sample, left_index=True, right_index=True
        )
        
    def fit_one_sample(self, 
                       seed: int,
                       all_space: Union[List, None],
                       all_time: Union[List, None],
                       space_subset: List[str],
                       time_subset: List[int],
                       strict: bool,
                       benefit_callable:Union[Callable, None],
                       cost_callable:Union[Callable, None],
                       cost_kwargs:dict,
                       benefit_kwargs:dict,
                       **kwargs) -> dict:
        """Draw one MonteCarlo sample and optimize. To be used with `fit_all_samples`. 
        
        `benefit_callable` and `cost_callable` must be functions with arguments `(seed, benefit_col, data)`. `cost_kwargs` and `benefit_kwargs` can be input to override defaults, such as if you want to use a different seed or change the name of a column.
        
        Note: `benefit_col` and `cost_col` for these callables denote the name of the resulting columns of the draw, not the original variable names.

        Args:
            seed (int): random seed
            all_space (Union[List, None]): spatial constraints (as in `Minimod`)
            all_time (Union[List, None]): time constraints (as in `Minimod`)
            space_subset (List[str]): subset for space constraints (as in `Minimod`)
            time_subset (List[int]): subset for time constraints (as in `Minimod`)
            strict (bool): whether to treat list of intervention names input *strictly* or using regex
            benefit_callable (Union[Callable, None]): The function for benefits transformation
            cost_callable (Union[Callable, None]): The function for cost transformation
            cost_kwargs (dict): extra arguments for the cost_callabe
            benefit_kwargs (dict): extra arguments for benefits callable

        Returns:
            dict: A dictionary of fitted results
        """    
        
        if benefit_callable is None:
            benefit_callable = self._construct_benefit_sample
        if cost_callable is None:
            cost_callable = self._construct_cost_sample

        cost_kwargs_default = {'seed' : seed, 'cost_col' : 'cost_random_draw', 'data' : self.data}
        benefit_kwargs_default = {'seed' : seed, 'benefit_col' : 'benefit_random_draw', 'data' : self.data}
        
        if cost_kwargs is not None:
            cost_kwargs_default.update(cost_kwargs)
        if benefit_kwargs is not None:
            benefit_kwargs_default.update(benefit_kwargs)

                       
        df = self._merge_samples(benefit_callable=benefit_callable,
                                 cost_callable=cost_callable,
                                 benefit_kwargs=benefit_kwargs_default,
                                 cost_kwargs=cost_kwargs_default) 

        minimod = Minimod(solver_type=self.solver_type)(
            data=df,
            intervention_col=self.intervention_col,
            space_col=self.space_col,
            time_col=self.time_col,
            benefit_col=benefit_kwargs_default.get('benefit_col'),
            cost_col=cost_kwargs_default.get('cost_col'),
            all_space=all_space,
            all_time=all_time,
            space_subset=space_subset,
            time_subset=time_subset,
            show_output=False,
            strict=strict,
            benefit_title="Effective Coverage",
            **kwargs,
        )

        minimod_opt.fit()
        
        # Run `minimod_opt.report` to get opt_df for iteration
        # Also save the opt_chosen dataframes in case there are multiple solutions
        
        opt_df_list = []
        
        for i in range(minimod_opt.num_solutions):
            minimod_opt.report(sol_num=i, quiet=True)
            opt_df_list.append(minimod_opt.opt_df)
            
        #TODO: add lowest cost per life saved index into iteration dict
            
        iteration_dict = {
            "status": minimod_opt.status,
            "opt_objective": [df['opt_costs_discounted'].sum() for df in opt_df_list],
            "opt_constraint": [df["opt_benefit_discounted"].sum() for df in opt_df_list],
            "num_vars": minimod_opt.num_cols,
            "constraints": minimod_opt.num_rows,
            "solutions": minimod_opt.num_solutions,
            "num_int": minimod_opt.num_int,
            "num_nz": minimod_opt.num_nz,
            "opt_df": opt_df_list,
            "sense" : minimod_opt.sense,
            "solver_name" : minimod_opt.solver_name,
            "minimum_benefit" : minimod_opt.minimum_benefit,
            "benefit_title" : minimod_opt.benefit_title,
            "bau_draw" : minimod_opt.bau_df
        }
        
        return iteration_dict

    def fit_all_samples(
        self,
        n_jobs = 5,
        N=None,
        all_space=None,
        all_time=None,
        space_subset=None,
        time_subset=None,
        strict=False,
        exception_behavior = 'immediate',
        only_optimal=False,
        benefit_callable=None,
        cost_callable=None,
        benefit_kwargs=None,
        cost_kwargs=None,
        random_seeds = None,
        **kwargs
    ):
        
        if N is None:
            N = 10

        sim_dict = {}


        print(f"""Running with {N} Samples""")
        
        partial_fit_sample = partial(self.fit_one_sample, 
                                     all_space=all_space,
                                     all_time=all_time,
                                     space_subset=space_subset,
                                     time_subset = time_subset,
                                     strict=strict,
                                     benefit_callable=benefit_callable,
                                     cost_callable=cost_callable,
                                     benefit_kwargs=benefit_kwargs,
                                     cost_kwargs=cost_kwargs,
                                     **kwargs)
        
        if random_seeds is None:
            random_seeds = range(N)
                
        sim_dict = pqdm(random_seeds, partial_fit_sample, n_jobs=n_jobs, exception_behaviour=exception_behavior)
        
        sim_df = pd.DataFrame(sim_dict)
        
        self.perc_opt = sim_df["status"].value_counts(normalize=True)[0] * 100

        if only_optimal:
            self.sim_results = sim_df.loc[lambda df: df['status'] == OptimizationStatus.OPTIMAL]
        else:
            self.sim_results = sim_df

        self.N = self.sim_results.shape[0]

        return self.sim_results

    def _all_opt_df(self, sol_filter=None):
        """Appends the dataframe from all simulation iterations together
        """
        #TODO: #28 Allow for concatenation of a combination of solutions, or all
        
        # First get sim_results so that `opt_df` is a series of dataframes
        # Turn list into numpy since it has a `take` method
        
        if sol_filter=='min_cb':
            # Find the solution with the highest benefit/cost ratio
            sol_num_all_opt_df = self.sim_results.assign(best_solution  = lambda df: df.apply(lambda df: (np.array(df['opt_objective'])/np.array(df['opt_constraint'])).argmin(), axis=1),
                                       new_opt_df = lambda df: df.apply(lambda x: x['opt_df'][x['best_solution']], axis=1))
        else:
            sol_num_all_opt_df = self.sim_results.assign(new_opt_df = lambda df: df['opt_df'].apply(lambda x: pd.concat(x)))
        
        all_opt_df = pd.concat(sol_num_all_opt_df.apply(lambda x: x['new_opt_df'].assign(iteration = x.name), axis=1).tolist())

        return all_opt_df

    def _get_intervention_group(self, data, intervention, strict=False):

        
        if strict:
            
            int_group = (
                data
                .loc[lambda df: df.index.
                    get_level_values(level= self.intervention_col)
                    .isin(intervention)]
            )
        
        else:
            int_group = (
                data
                .loc[lambda df: df.index.
                    get_level_values(level= self.intervention_col)
                    .str.contains(intervention)]
            )

        return int_group
    
    def _get_indicator_if_in_intervention(self, name, indicator_spec = None, strict=False):
        
        if indicator_spec is None:
            indicator_spec = 1
        
        return (self._get_intervention_group(self._all_opt_df(), name, strict=strict)
                .reset_index()
                [['opt_vals', 'iteration', self.intervention_col, self.space_col, self.time_col]]
                .groupby('iteration')
                .agg(lambda x: 1 if x.sum() > indicator_spec else 0)
                )

    def report(
        self,
        avg_time=False,
        avg_space=False,
        intervention_group=None,
        indicator_spec = None,
        strict=False
    ):

        avg = self.sim_results.convert_dtypes().mean()

        s = OptimizationSummary(self)

        header = [
            ("MiniMod Solver Results", ""),
            ("Method:", str(self.sim_results['sense'].min())),
            ("Solver:", str(self.sim_results['solver_name'].min())),
            ("Percentage Optimized:", self.perc_opt),
            ("Average Number Solutions Found:", avg["solutions"]),
        ]

        features = [
            ("No. of Variables:", avg["num_vars"]),
            ("No. of Integer Variables:", avg["num_int"]),
            ("No. of Constraints", avg["constraints"]),
            ("No. of Non-zeros in Constr.", avg["num_nz"]),
        ]

        results_benefits = [("Minimum Benefit", self.sim_results.minimum_benefit.mean())]

        stats = [
            ("Statistics for Benefits and Costs", ""),
        ]

        s.print_generic(header, features, results_benefits, stats)

        stats_df = (
            self.sim_results[["opt_objective", "opt_constraint"]]
            .astype(float)
            .describe()
            .round(4)
            .to_markdown()
        )
        print(stats_df)

        if intervention_group is not None:

            s.print_generic([(f"% Appearance of:", "")])

            for i in intervention_group:

                int_group = (self._get_indicator_if_in_intervention(i, 
                                                                    indicator_spec=indicator_spec,
                                                                    strict=strict).sum()/self.N*100)['opt_vals']

                s.print_generic([(f"{i}", f"{int_group}")])

        if avg_time:

            time_df = (
                self._all_opt_df().groupby([self.time_col, "iteration"])
                .sum()
                .groupby(self.time_col)
                .mean()[["opt_benefit", "opt_costs"]]
            )

            s.print_generic([("Mean Benefits and Costs across time", "")])
            print(time_df.to_markdown())

        if avg_space:

            space_df = (
                self._all_opt_df().groupby([self.space_col, "iteration"])
                .sum()
                .groupby(self.space_col)
                .mean()[["opt_benefit", "opt_costs"]]
            )

            s.print_generic([("Mean Benefits and Costs across Regions", "")])
            print(space_df.to_markdown())

    def plot_opt_hist(self, save=None):

        p = Plotter(self)

        costs = "Optimal Costs"
        benefits = self.sim_results['benefit_title'].min()

        if self.solver_type == "costmin":

            objective_title = costs
            constraint_title = benefits

        elif self.solver_type == "benmax":

            objective_title = benefits
            constraint_title = costs


        self.sim_results['opt_constraint2']=self.sim_results['opt_constraint'].apply(lambda x: x[0]/1000)
        self.sim_results['opt_objective2']=self.sim_results['opt_objective'].apply(lambda x: x[0]/1000)

        fig, (benefit_plot, cost_plot) = p._plot_sim_hist(
            data=self.sim_results,
            benefit_col="opt_constraint2",
            cost_col="opt_objective2",
            #benefit_col="opt_constraint",
            #cost_col="opt_objective",
            cost_title=objective_title,
            benefit_title=constraint_title,
            save=save,
        )

        benefit_plot.xaxis.set_major_formatter(mtick.StrMethodFormatter('{x:,.0f}'))
        cost_plot.xaxis.set_major_formatter(mtick.StrMethodFormatter('{x:,.0f}'))
    
        benefit_plot.set_xlabel("Thousands of Individuals")
        cost_plot.set_xlabel("Thousands of 2019 USD")

        benefit_xlims = benefit_plot.get_xlim()
        benefit_ylims = benefit_plot.get_ylim()

        # Put text at midpoint of y
        text_y = (benefit_ylims[1] - benefit_ylims[0]) / 2

        # offset by 10% of length of x-axis
        text_x = (
            self.sim_results['minimum_benefit'].mean()/1000 + (benefit_xlims[1] - benefit_xlims[0]) * 0.1
        )

        benefit_plot.axvline(self.sim_results['minimum_benefit'].mean()/1000, color="red")
        benefit_plot.text(text_x, text_y, "Mean\nMinimum\nBenefit\nConstraint")
        
        # Get total cost for a draw 
        cost_xlims = cost_plot.get_xlim()
        cost_ylims = cost_plot.get_ylim()

        # Put text at midpoint of y
        cost_plot.axvline(self.sim_results['bau_draw'].apply(lambda x: x['discounted_costs'].sum()).mean()/1000, color='red')

        text_y2 = (cost_ylims[1] - cost_ylims[0]) / 2

        # offset by 10% of length of x-axis
        text_x2 = (
            self.sim_results['bau_draw'].apply(lambda x: x['discounted_costs'].sum()).mean()/1000 + (cost_xlims[1] - cost_xlims[0]) * 0.1
        )
        cost_plot.text(text_x2, text_y2, "Mean\nBAU\nCost")

        return fig, (benefit_plot, cost_plot)

    def plot_sim_trajectories(self, data_of_interest="benefits", save=None):

        fig, ax = plt.subplots()

        if data_of_interest == "benefits":
            col_of_interest = "opt_benefit"
            ylabel_interest = "Individuals"
        elif data_of_interest == "costs":
            col_of_interest = "opt_costs"
            ylabel_interest = "2019 USD"
        
        df_all = self.sim_results['opt_df'].apply(lambda x: x[col_of_interest].groupby(self.time_col).sum()).T

        # Now get mean trajectory

        df_all.plot(color='red', alpha=0.09, ax=ax, legend=False)
        df_all.mean(axis=1).plot(ax=ax, color="black")

        # plt.figtext(0, 0, "Bold line represents mean trajectory.")
        ax.set_title("Trajectories of all Simulations")
        
        ax.yaxis.set_major_formatter(mtick.StrMethodFormatter('{x:,.0f}'))
        ax.set_ylabel(ylabel_interest)

        if save is not None:
            plt.savefig(save, dpi=160)

        return ax

    def plot_intervention_stacked(self, intervention_group=None, intervention_names = None, indicator_spec=3):
        
        fig, ax = plt.subplots()

        all_opt_df = (self._all_opt_df()
                      .groupby(['intervention', 'time', 'iteration'])
                      ['opt_vals']
                      .sum()
                      .to_frame()
                      .assign(opt_vals = lambda df: (df['opt_vals']>indicator_spec).astype(int))
                      )
        
        int_group = (
            all_opt_df[all_opt_df['opt_vals']>indicator_spec]
            .reset_index(level=self.intervention_col)
            [self.intervention_col]
            .str.extractall('|'.join([f"(?P<{j}>{i})" for i, j in zip(intervention_group, intervention_names)]))
         )             
        
        int_group.groupby(self.time_col).count().apply(lambda x: x/x.sum(), axis=1).plot.bar(stacked=True, ax=ax)
                   
        
        ax.legend(loc = 'lower left', bbox_to_anchor=(1.0, 0.5))
        ax.set_ylabel("% of Occurrences")
        ax.set_xlabel("Time")
        
        return ax

    


import mip
import pandas as pd

from minimod_opt.utils.suppress_messages import suppress_stdout_stderr

from minimod_opt.utils.exceptions import NoVars, NoObjective, NoConstraints
from itertools import combinations, permutations


class Model:


    def __init__(self, data:pd.DataFrame, sense:str, solver_name:str, show_output:bool):
        """A class that instantiates a `mip` model.

        Args:
            data (pd.DataFrame): The input dataframe that includes benefit/cost data
            sense (str): Whether the model should minimize (MINIMIZE) or maximize (MAXIMIZE)
            solver_name (str): The solver type (CBC or some other mip solver)
            show_output (bool):  Whether to show output of model construction
        """
       

        ## Tell the fitter whether to maximize or minimize
        self.model = mip.Model(sense=sense, solver_name=solver_name)

        self._df = data

        self.show_output = show_output

        if self.show_output:
            self.model.verbose = 1
        else:
            self.model.verbose = 0
        ## set tolerances based on GAMS tolerances
        # primal tol -> infeas
        # dual tol -> opt tol
        # integer  tol -> integer_tol
        # self.model.opt_tol = 1e-10
        # self.model.infeas_tol = 1e-10
        # self.model.integer_tol = 1e-06
        # self.model.seed = 0
        # # self.model.clique = 0

        # # # # # allowable gap/ optca -> max_mip_gap_abs
        # # # # # ratioGap/ optcr -> max_mip_gap

        # self.model.max_mip_gap_abs = 0
        # self.model.max_mip_gap = 0.1

        # self.model.preprocess = 0
        # self.model.lp_method = -1
        # self.cut_passes = 1

    def _stringify_tuple(self, tup:tuple)->str:
        """This function converts all the items of a tuple into one string

        Args:
            tup (tuple): general tuple 

        Returns:
            str: output string generated from all items in the tuple
        """

        strings = [str(x) for x in tup]

        stringified = "".join(strings)

        return stringified

    def _model_var_create(self):

        self._df["mip_vars"] = self._df.apply(
            lambda x: self.model.add_var(
                var_type=mip.BINARY, name=f"x_{self._stringify_tuple(x.name)}"
            ),
            axis=1,
        )

    def _base_constraint(self, space:str, time:str):
        """Constraint making the 'ones' constraint, making an intervention a binary intervention.

        Args:
            space (str): name of dataframe's column with information on regions/locations
            time (str): name of dataframe's column with information on time/years
        """
        grouped_df = self._df["mip_vars"].groupby([space, time]).agg(mip.xsum)

        base_constrs = grouped_df.values

        # Get constraint name
        names = list(
            map(
                self._stringify_tuple,
                grouped_df.index.to_list()
            )
        )

        for constr, name in zip(base_constrs, names):

            self.model += constr <= 1, "ones_" + name

    def _intervention_subset(self, intervention:str, strict:bool, subset_names:list =[])-> dict:
        """This function creates a dictionary for the subset of specified interventions

        Args:
            intervention (str): name of dataframe's column with set of interventions 
            strict (bool): whether to look for specific intervention names (using DataFrame.isin) or whether to look for a regex (using DataFrame.str.contains)
            subset_names (list, optional): Which intervention names to look for. Defaults to [].

        Returns:
            dict: dictionary with subset of interventions
        """

        subset_dict = {}

        for i in subset_names:

            if strict:
                subset_dict[i[0]] = self._df.loc[
                    lambda df: df.index.get_level_values(level=intervention).isin([i])
                ]

                if subset_dict[i[0]].empty:
                    raise Exception(f"'{i[0]}' not found in dataset.")

            else:
                subset_dict[i] = self._df.loc[
                    lambda df: df.index.get_level_values(
                        level=intervention
                    ).str.contains(i, case=False)
                ]

                if subset_dict[i].empty:
                    raise Exception(f"'{i}' not found in dataset.")

        return subset_dict

    def _all_constraint(
        self,
        strict:bool,
        intervention:str=None,
        space:str=None,
        time:str=None,
        subset_names:list=None,
        over:str=None,
        subset_list:list=None,
    ):
        """A constraint across space or time, where an intervention is tied to other interventions across space time (for a given subset). For instance, if you wanted to
        create a national intervention, for some intervention X, with 2 regions, this constraint would do the following:
        
        $X_{region1,t} == X_{region2,t}$ -> _all_constraint(over='space', subset_names=None)
        
        If you wanted to make sure that interventions were tied across time so that period 2 could only be used if period 1 was used:
        
        $X_{region1, 1} == X_{region1, 2}$
       
        Args:
            strict (bool): whether to look for specific intervention names (using DataFrame.isin) or whether to look for a regex (using DataFrame.str.contains)
            intervention (str, optional): name of dataframe's column with set of interventions. Defaults to None.
            space (str, optional): name of dataframe's column with information on regions/locations. Defaults to None.
            time (str, optional): name of dataframe's column with information on time/year. Defaults to None.
            subset_names (list, optional): Which intervention names to look for. Defaults to None.
            over (str, optional): name of dataframe's column  with attribute used to group data by (e.g., time, region). Defaults to None.
            subset_list (list, optional): A list of interventions that are being constrained. Defaults to None.
        """

        subset_dict = self._intervention_subset(
            intervention=intervention, strict=strict, subset_names=subset_names
        )
        for sub in subset_dict.keys():

            mip_vars_grouped_sum = (
                subset_dict[sub].groupby([space, time])["mip_vars"].agg(mip.xsum)
            )

            if over == time:
                slicer = space
            elif over == space:
                slicer = time

            unstacked = mip_vars_grouped_sum.unstack(level=slicer)

            if subset_list is None:
                subset_list = unstacked.index

            # get combinations of different choices
            constraint_combinations = permutations(unstacked.index.tolist(), 2)

            constraint_list = [
                (i, j) for (i, j) in constraint_combinations if i in subset_list
            ]

            for col in unstacked.columns:
                for (comb1, comb2) in constraint_list:
                    self.add_constraint(
                        unstacked[col].loc[comb1], unstacked[col].loc[comb2], 
                        "eq",
                        name = str(sub) + "_" + str(col) + "_" + str(comb1) + "_" + str(comb2)
                    )

    def _all_space_constraint(
        self,
        strict:bool,
        intervention:str=None,
        space:str=None,
        time:str=None,
        subset_names:list=None,
        over:str=None,
        subset_list:list=None,
    )->None:
        """This function invokes the function `_all_constraint' to create national interventions

        Args:
            strict (bool): whether to look for specific intervention names (using DataFrame.isin) or whether to look for a regex (using DataFrame.str.contains)
            intervention (str, optional): name of dataframe's column with set of interventions. Defaults to None.
            space (str, optional): name of dataframe's column with informaiton on regions/locations. Defaults to None.
            time (str, optional): name of dataframe's column with information on time/year. Defaults to None.
            subset_names (list, optional): Which intervention names to look for. Defaults to None.
            over (str, optional): name of dataframe's column  with attribute used to group data by (e.g., time, region). Defaults to None.
            subset_list (list, optional): A list of interventions that are being constrained. Defaults to None.
        """

        return self._all_constraint(
            strict,
            intervention=intervention,
            space=space,
            time=time,
            subset_names=subset_names,
            over=over,
            subset_list=subset_list,
        )

    def _all_time_constraint(
        self,
        strict:bool,
        intervention:str=None,
        space:str=None,
        time:str=None,
        subset_names:list=None,
        over:str=None,
        subset_list:list=None,
    )->None:
        """This function invokes the function `_all_constraint'

        Args:
            strict (bool): whether to look for specific intervention names (using DataFrame.isin) or whether to look for a regex (using DataFrame.str.contains)
            intervention (str, optional): name of dataframe's column with set of interventions. Defaults to None.
            space (str, optional): name of dataframe's column with information on regions/locations. Defaults to None.
            time (str, optional): name of dataframe's column with information on time/year. Defaults to None.
            subset_names (list, optional): Which intervention names to look for. Defaults to None.
            over (str, optional): name of dataframe's column  with attribute used to group data by (e.g., time, region). Defaults to None.
            subset_list (list, optional): A list of interventions that are being constrained. Defaults to None.
        """

        return self._all_constraint(
            strict,
            intervention=intervention,
            space=space,
            time=time,
            subset_names=subset_names,
            over=over,
            subset_list=subset_list,
        )

    
    def get_equation(self, name:str=None, show:bool=True)->str or mip.LinExpr or mip.Constr :
        """This function returns a constraint by its name. If no name is specified, returns all constraints

        Args:
            name (str, optional): a string corresponding to the name of the constraint. Defaults to None.
            show (bool, optional): whether to return the contraint in string type or not. Defaults to True.

        Returns:
            str or mip.LinExpr or mip.Constr: A `mip` constraint object
        """

        if name is None:
            return self.model.constrs
        elif name == 'objective':
            if show:
                return str(self.model.objective)
            else:
                return self.model.objective
        else:
            if show:
                return str(self.model.constr_by_name(name))
            else:
                return self.model.constr_by_name(name)

    def add_objective(self, eq:mip.LinExpr):
        """Sets the objective function of the problem as a linear expression

        Args:
            eq (mip.LinExpr): equation defining the objective function
        """

        self.model.objective = eq

    def add_constraint(self, eq:mip.LinExpr, constraint:mip.LinExpr, way:str="ge", name:str=""):
        """This function merges the objective function of the model with its constraint

        Args:
            eq (mip.LinExpr): equation defining the objective function
            constraint (mip.LinExpr):equation defining the constraint function
            way (str, optional): whether greater or equal (ge), less or equal(le), or equal(eq). Defaults to "ge".
            name (str, optional): optional name for the constraint. Defaults to "".
        """
        
        if isinstance(constraint, pd.Series):
            # Merge equation with constraint
            df = eq.merge(constraint, left_index = True, right_index= True)
            
            for i, ee, c in df.itertuples():
                if way == "ge":
                    self.model += ee >= c, name
                elif way == "le":
                    self.model += ee <= c, name
                elif way == "eq":
                    self.model += ee == c, name
                
            
        else:
            if way == "ge":
                self.model += eq >= constraint, name
            elif way == "le":
                self.model += eq <= constraint, name
            elif way == "eq":
                self.model += eq == constraint, name

    def base_model_create(
        self,
        intervention:str,
        space:str,
        time:str,
        all_time:list=None,
        all_space:list=None,
        time_subset:list=None,
        space_subset:list=None,
        strict:bool=False,
    ):
        """A function   

        Args:
            intervention (str): name of dataframe's column with set of interventions
            space (str): name of dataframe's column with information on regions/locations
            time (str): name of dataframe's column with information on time/years
            all_time (list, optional): list of intervention vehicles that are active during all periods (e.g., cube, oil). Defaults to None.
            all_space (list, optional): list of intervention vehicles that are targeted at a country-wide level (e.g., cube, oil). Defaults to None.
            time_subset (list, optional): list with subset of periods. Defaults to None.
            space_subset (list, optional):list with subset of regions/locations. Defaults to None.
            strict (bool, optional): whether to look for specific intervention names (using DataFrame.isin) or whether to look for a regex (using DataFrame.str.contains). Defaults to False.

        """

        ## Now we create the choice variable, x, which is binary and is the size of the dataset.
        ## In this case, it should just be a column vector with the rows equal to the data:

        self._model_var_create()

        ## First add base constraint, which only allows one intervention per time and space
        self._base_constraint(space, time)

        ## Add all_space or all_time constraints if necessary
        if all_time is not None:

            if intervention is None or space is None:
                raise Exception("One of the subset columns were not found")

            self._all_time_constraint(
                strict,
                intervention=intervention,
                space=space,
                time=time,
                subset_names=all_time,
                over=time,
                subset_list=time_subset,
            )

        if all_space is not None:

            if intervention is None or time is None:
                raise Exception("One of the subset columns were not found")

            self._all_space_constraint(
                strict,
                intervention=intervention,
                space=space,
                time=time,
                subset_names=all_space,
                over=space,
                subset_list=space_subset,
            )

    def optimize(self, **kwargs):
        """This function conducts the optimization procedure 

        Args:
            **kwargs: Other parameters for optimization procedure using `mip.optimize' (max_seconds, max_nodes, max_solutions)
        """

        self.status = None

        if self.model.num_cols == 0:
            raise NoVars("No Variables added to the model")
        if self.model.num_rows == 0:
            raise NoConstraints("No constraints added to the model.")
        try:
            self.model.objective
        except Exception:
            raise NoObjective("No Objective added to the model")

        # Now, allow for arguments to the optimize function to be given:

        max_seconds = kwargs.pop("max_seconds", mip.INF)
        max_nodes = kwargs.pop("max_nodes", mip.INF)
        max_solutions = kwargs.pop("max_solutions", mip.INF)

        if self.show_output:
            self.status = self.model.optimize(max_seconds, max_nodes, max_solutions)
        else:
            with suppress_stdout_stderr():
                self.status = self.model.optimize(max_seconds, max_nodes, max_solutions)
        if self.show_output:
            if self.status == mip.OptimizationStatus.OPTIMAL:
                print("[Note]: Optimal Solution Found")
            elif self.status == mip.OptimizationStatus.FEASIBLE:
                print("[Note]: Feasible Solution Found. This may not be optimal.")
            elif self.status == mip.OptimizationStatus.NO_SOLUTION_FOUND:
                print("[Warning]: No Solution Found")
            elif self.status == mip.OptimizationStatus.INFEASIBLE:
                print("[Warning]: Infeasible Solution Found")

    def process_results(self, benefit_col:str, cost_col:str, intervention_col:str, space_col:str, sol_num:int = None)->pd.DataFrame:
        """This function creates a dataframe with information on benefits and costs for the optimal interventions

        Args:
            benefit_col (str): name of dataframe's column with benefits
            cost_col (str): name of dataframe's column with costs
            intervention_col (str): name of dataframe's column with set of interventions
            space_col (str): name of dataframe's column with information on regions/locations
            sol_num (int, optional): index of solution. Defaults to None.

        Returns:
            pd.DataFrame: dataframe with optimal interventions
        """

        
        if isinstance(sol_num, int):
            opt_df = self._df.copy(deep=True).assign(opt_vals=lambda df: df["mip_vars"].apply(lambda y: y.xi(sol_num)))
        else:
            opt_df = self._df.copy(deep=True).assign(
                opt_vals=lambda df: df["mip_vars"].apply(lambda y: y.x))

        opt_df = (opt_df.assign(
            opt_benefit=lambda df: df[benefit_col] * df["opt_vals"],
            opt_costs=lambda df: df[cost_col] * df["opt_vals"],
            opt_costs_discounted=lambda df: df["discounted_costs"] * df["opt_vals"],
            opt_benefit_discounted=lambda df: df["discounted_benefits"]* df["opt_vals"])
                  .infer_objects()
                  .assign(
            cumulative_discounted_benefits = lambda df: (df
                                                         .groupby([space_col])['opt_benefit_discounted']
                                                         .transform('cumsum')),
            cumulative_discounted_costs = lambda df: (df
                                                         .groupby([space_col])['opt_costs_discounted']
                                                         .transform('cumsum')),
            cumulative_benefits = lambda df: (df
                                                         .groupby([space_col])['opt_benefit']
                                                         .transform('cumsum')),
            cumulative_costs = lambda df: (df
                                                         .groupby([space_col])['opt_costs']
                                                         .transform('cumsum'))
        )[
            [
                "opt_vals",
                "opt_benefit",
                "opt_costs",
                "opt_costs_discounted",
                "opt_benefit_discounted",
                "cumulative_discounted_benefits",
                "cumulative_discounted_costs",
                "cumulative_benefits",
                "cumulative_costs"
                ]
        ])
    

        return opt_df

    def write(self, filename:str="model.lp"):
        """Thi function saves model to file

        Args:
            filename (str, optional): name of file. Defaults to "model.lp".
        """
        self.model.write(filename)

    def get_model_results(self)->list:
        """This function returns a list with additional results of the optimization procedure

        Returns:
            list: model's results
        """

        return (
            self.model.objective_value,
            self.model.objective_values,
            self.model.objective_bound,
            self.model.num_solutions,
            self.model.num_cols,
            self.model.num_rows,
            self.model.num_int,
            self.model.num_nz,
            self.status,
        )

